#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.02.02 02:00:00                  #
# ================================================== #

from typing import Optional

from pygpt_net.core.events import Event, AppEvent
from pygpt_net.item.model import ModelItem

from .editor import Editor


class Model:
    def __init__(self, window=None):
        """
        Model controller

        :param window: Window instance
        """
        self.window = window
        self.editor = Editor(window)

    def select(self, model: str):
        """
        Select model

        :param model: model ID
        """
        # check if model change is not locked
        if self.change_locked():
            return

        mode = self.window.core.config.get('mode')
        self.window.core.config.set('model', model)
        if 'current_model' not in self.window.core.config.data:
            self.window.core.config.data['current_model'] = {}
        self.window.core.config.data['current_model'][mode] = model

        event = Event(Event.MODEL_SELECT, {
            'value': model,
        })
        self.window.dispatch(event)

        # update all layout
        self.window.controller.ui.update()
        self.window.dispatch(AppEvent(AppEvent.MODEL_SELECTED))  # app event

    def next(self):
        """Select next model"""
        mode = self.window.core.config.get('mode')
        model = self.window.core.config.get('model')
        next = self.window.core.models.get_next(model, mode)
        self.select(next)

    def prev(self):
        """Select previous model"""
        mode = self.window.core.config.get('mode')
        model = self.window.core.config.get('model')
        prev = self.window.core.models.get_prev(model, mode)
        self.select(prev)

    def set(self, mode: str, model: str):
        """
        Set model by mode and model name

        :param mode: mode name
        :param model: model name
        """
        self.window.core.config.set('model', model)
        if 'current_model' not in self.window.core.config.data:
            self.window.core.config.data['current_model'] = {}
        self.window.core.config.data['current_model'][mode] = model

    def set_by_idx(self, mode: str, idx: int):
        """
        Set model by mode and model idx

        :param mode: mode name
        :param idx: model index
        """
        model = self.window.core.models.get_by_idx(idx, mode)
        self.window.core.config.set('model', model)
        if 'current_model' not in self.window.core.config.data:
            self.window.core.config.data['current_model'] = {}
        self.window.core.config.data['current_model'][mode] = model

        event = Event(Event.MODEL_SELECT, {
            'value': model,
        })
        self.window.dispatch(event)

    def select_on_list(self, model: str):
        """
        Select model on list

        :param model: model ID
        """
        self.window.ui.nodes["prompt.model"].set_value(model)

    def select_current(self):
        """Select current model on list"""
        mode = self.window.core.config.get('mode')
        model = self.window.core.config.get('model')
        items = self.window.core.models.get_by_mode(mode)
        if model in items:
            self.select_on_list(model)

    def select_default(self):
        """Set default model"""
        model = self.window.core.config.get('model')
        if model is None or model == "":
            mode = self.window.core.config.get('mode')

            # set previous selected model
            current_models = self.window.core.config.get('current_model')
            if mode in current_models and \
                    current_models[mode] is not None and \
                    current_models[mode] != "" and \
                    current_models[mode] in self.window.core.models.get_by_mode(mode):
                self.window.core.config.set('model', current_models[mode])
            else:
                # or set default model
                self.window.core.config.set('model', self.window.core.models.get_default(mode))

    def switch_inline(
            self,
            mode: str,
            model: Optional[ModelItem] = None
    ) -> ModelItem:
        """
        Switch inline model instance (force change model if needed)

        :param mode: mode (vision, chat, etc.)
        :param model: model instance
        :return: model instance
        """
        event = Event(Event.MODEL_BEFORE, {
            'mode': mode,
            'model': model,  # model instance
        })
        self.window.dispatch(event)
        tmp_model = event.data['model']
        if tmp_model is not None:
            model = tmp_model
        return model

    def init_list(self):
        """Init models list"""
        mode = self.window.core.config.get('mode')
        items = {}
        data = self.window.core.models.get_by_mode(mode)
        for k in data:
            suffix = ""
            if "provider" in data[k].llama_index and data[k].llama_index["provider"] == "ollama":
                suffix = " [Ollama]"
            items[k] = data[k].name + suffix
        items = dict(sorted(items.items(), key=lambda item: item[1]))  # sort items by name
        self.window.ui.nodes["prompt.model"].set_keys(items)

    def update(self):
        """Update models"""
        self.select_default()
        self.select_current()

    def change_locked(self) -> bool:
        """
        Check if model change is locked

        :return: True if locked
        """
        return self.window.controller.chat.input.generating



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.21 20:00:00                  #
# ================================================== #

class AgentDebug:
    def __init__(self, window=None):
        """
        Agent debug

        :param window: Window instance
        """
        self.window = window
        self.id = 'agent'

    def update(self):
        """Update debug window"""
        self.window.core.debug.begin(self.id)
        self.window.core.debug.add(self.id, '[Llama-index]', '')
        self.window.core.debug.add(self.id, 'agents', str(self.window.core.agents.provider.get_ids()))
        self.window.core.debug.add(self.id, 'eval_step', str(self.window.controller.agent.llama.eval_step))
        self.window.core.debug.add(self.id, '[LEGACY]', '')
        self.window.core.debug.add(self.id, 'iteration', str(self.window.controller.agent.legacy.iteration))
        self.window.core.debug.add(self.id, 'limit', str(self.window.core.config.get("agent.iterations")))
        self.window.core.debug.add(self.id, 'prev_output', str(self.window.controller.agent.legacy.prev_output))
        self.window.core.debug.add(self.id, 'is_user', str(self.window.controller.agent.legacy.is_user))
        self.window.core.debug.add(self.id, 'stop', str(self.window.controller.agent.legacy.stop))
        self.window.core.debug.add(self.id, 'finished', str(self.window.controller.agent.legacy.finished))
        self.window.core.debug.add(self.id, 'allowed_cmds', str(self.window.controller.agent.legacy.allowed_cmds))

        self.window.core.debug.end(self.id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.18 21:00:00                  #
# ================================================== #

from PySide6.QtCore import Slot, Signal
from pygpt_net.plugin.base.worker import BaseWorker, BaseSignals

class WorkerSignals(BaseSignals):
    screenshot = Signal(dict, object)

class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.window = None
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.cmds = None
        self.ctx = None
        self.msg = None

    @Slot()
    def run(self):
        responses = []
        for item in self.cmds:
            if self.is_stopped():
                break
            response = None
            try:
                if item["cmd"] in self.plugin.allowed_cmds and self.plugin.has_cmd(item["cmd"]):

                    if item["cmd"] == "camera_capture":
                        response = self.cmd_camera_capture(item)

                    elif item["cmd"] == "make_screenshot":
                        response = self.cmd_make_screenshot(item)

                    elif item["cmd"] == "analyze_image_attachment":
                        response = self.cmd_analyze_image_attachment(item)

                    elif item["cmd"] == "analyze_screenshot":
                        response = self.cmd_analyze_screenshot(item)

                    elif item["cmd"] == "analyze_camera_capture":
                        response = self.cmd_analyze_camera_capture(item)

                    if response:
                        responses.append(response)

            except Exception as e:
                responses.append(
                    self.make_response(
                        item,
                        self.throw_error(e)
                    )
                )

        # send response
        if len(responses) > 0:
            self.reply_more(responses)

    def cmd_camera_capture(self, item: dict) -> dict:
        """
        Capture image from camera

        :param item: command item
        :return: response item
        """
        try:
            if self.window.controller.camera.internal_capture():
                result = "OK"
            else:
                result = "FAILED"
        except Exception as e:
            result = self.throw_error(e)
        return self.make_response(item, result)

    def cmd_make_screenshot(self, item: dict) -> dict:
        """
        Make desktop screenshot

        :param item: command item
        :return: response item
        """
        try:
            self.window.controller.painter.capture.screenshot()
            result = "OK"
        except Exception as e:
            result = self.throw_error(e)
        return self.make_response(item, result)

    def cmd_analyze_image_attachment(self, item: dict) -> dict:
        """
        Analyze image attachment

        :param item: command item
        :return: response item
        """
        try:
            prompt = ""
            if self.has_param(item, "prompt"):
                prompt = self.get_param(item, "prompt")
            path = ""
            if self.has_param(item, "path"):
                path = self.get_param(item, "path")
            result = self.window.core.vision.analyzer.from_path(
                ctx=self.ctx,
                prompt=prompt,
                path=path,
            )
        except Exception as e:
            result = self.throw_error(e)
        return self.make_response(item, result)

    def cmd_analyze_screenshot(self, item: dict) -> dict:
        """
        Analyze screenshot

        :param item: command item
        :return: response item
        """
        try:
            prompt = ""
            if self.has_param(item, "prompt"):
                prompt = self.get_param(item, "prompt")
            result = self.window.core.vision.analyzer.from_screenshot(
                ctx=self.ctx,
                prompt=prompt,
            )
        except Exception as e:
            result = self.throw_error(e)
        return self.make_response(item, result)

    def cmd_analyze_camera_capture(self, item: dict) -> dict:
        """
        Analyze camera capture

        :param item: command item
        :return: response item
        """
        try:
            prompt = ""
            if self.has_param(item, "prompt"):
                prompt = self.get_param(item, "prompt")
            result = self.window.core.vision.analyzer.from_camera(
                ctx=self.ctx,
                prompt=prompt,
            )
        except Exception as e:
            result = self.throw_error(e)
        return self.make_response(item, result)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.16 01:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QVBoxLayout, QLabel, QHBoxLayout, QCheckBox, QWidget, QSizePolicy, QPushButton, \
    QGridLayout, QWidgetItem, QLayout

from pygpt_net.ui.widget.audio.output import AudioOutput
from pygpt_net.ui.widget.element.labels import ChatStatusLabel, IconLabel, HelpLabel
from pygpt_net.ui.widget.anims.loader import Loading
from pygpt_net.ui.widget.tabs.layout import OutputLayout

from .explorer import Explorer
from .input import Input
from .calendar import Calendar
from .painter import Painter

from pygpt_net.utils import trans
import pygpt_net.icons_rc



class Output:
    def __init__(self, window=None):
        """
        Chat output UI

        :param window: Window instance
        """
        self.window = window
        self.explorer = Explorer(window)
        self.input = Input(window)
        self.calendar = Calendar(window)
        self.painter = Painter(window)

    def setup(self) -> QWidget:
        """
        Setup output

        :return: QWidget
        """
        # prepare columns
        self.window.ui.layout = OutputLayout(self.window)

        # create empty containers for chat output
        self.window.ui.nodes['output'] = {}
        self.window.ui.nodes['output_plain'] = {}

        layout = QVBoxLayout()
        layout.addWidget(self.window.ui.layout)
        layout.addLayout(self.setup_bottom())
        layout.setContentsMargins(0, 5, 0, 0)

        widget = QWidget()
        widget.setLayout(layout)
        return widget

    def setup_bottom(self) -> QHBoxLayout:
        """
        Setup bottom status bar

        :return: QHBoxLayout
        """
        # video capture icon
        self.window.ui.nodes['icon.video.capture'] = IconLabel(":/icons/webcam.svg")
        self.window.ui.nodes['icon.video.capture'].setToolTip(trans("icon.video.capture"))
        self.window.ui.nodes['icon.video.capture'].clicked.connect(
            lambda:self.window.controller.camera.toggle_capture()
        )

        # audio output icon
        self.window.ui.nodes['icon.audio.output'] = IconLabel(":/icons/volume.svg")
        self.window.ui.nodes['icon.audio.output'].setToolTip(trans("icon.audio.output"))
        self.window.ui.nodes['icon.audio.output'].clicked.connect(
            lambda: self.window.controller.plugins.toggle('audio_output')
        )

        # audio input icon
        self.window.ui.nodes['icon.audio.input'] = IconLabel(":/icons/mic.svg")
        self.window.ui.nodes['icon.audio.input'].setToolTip(trans("icon.audio.input"))
        self.window.ui.nodes['icon.audio.input'].clicked.connect(
            lambda: self.window.controller.plugins.toggle('audio_input')
        )

        # interpreter icon
        self.window.ui.nodes['icon.interpreter'] = IconLabel(":/icons/code.svg")
        self.window.ui.nodes['icon.interpreter'].setToolTip("Python Code Interpreter")
        self.window.ui.nodes['icon.interpreter'].clicked.connect(
            lambda: self.window.tools.get("interpreter").toggle()
        )

        # indexer icon
        self.window.ui.nodes['icon.indexer'] = IconLabel(":/icons/db.svg")
        self.window.ui.nodes['icon.indexer'].setToolTip("Indexer")
        self.window.ui.nodes['icon.indexer'].clicked.connect(
            lambda: self.window.tools.get("indexer").toggle()
        )

        # mode
        self.window.ui.nodes['chat.label'] = ChatStatusLabel("")
        self.window.ui.nodes['chat.label'].setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        self.window.ui.nodes['chat.label'].setWordWrap(False)

        # model
        self.window.ui.nodes['chat.model'] = ChatStatusLabel("")
        self.window.ui.nodes['chat.model'].setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        self.window.ui.nodes['chat.model'].setWordWrap(False)

        # plugins
        self.window.ui.nodes['chat.plugins'] = ChatStatusLabel("")
        self.window.ui.nodes['chat.plugins'].setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

        # timestamp
        self.window.ui.nodes['output.timestamp'] = QCheckBox(trans('output.timestamp'))
        self.window.ui.nodes['output.timestamp'].stateChanged.connect(
            lambda: self.window.controller.chat.common.toggle_timestamp(
                self.window.ui.nodes['output.timestamp'].isChecked())
        )

        # plain text
        self.window.ui.nodes['output.raw'] = QCheckBox(trans('output.raw'))
        self.window.ui.nodes['output.raw'].clicked.connect(
            lambda: self.window.controller.chat.common.toggle_raw(
                self.window.ui.nodes['output.raw'].isChecked())
        )

        """
        # edit icons
        self.window.ui.nodes['output.edit'] = QCheckBox(trans('output.edit'))
        self.window.ui.nodes['output.edit'].clicked.connect(
            lambda: self.window.controller.chat.common.toggle_edit_icons(
                self.window.ui.nodes['output.edit'].isChecked())
        )
        """

        # tokens
        self.window.ui.nodes['prompt.context'] = ChatStatusLabel("")
        self.window.ui.nodes['prompt.context'].setToolTip(trans('tip.tokens.ctx'))
        self.window.ui.nodes['prompt.context'].setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

        self.window.ui.nodes['input.counter'] = ChatStatusLabel("")
        self.window.ui.nodes['input.counter'].setToolTip(trans('tip.tokens.input'))
        self.window.ui.nodes['input.counter'].setWordWrap(False)

        # plugin audio output addon
        self.window.ui.plugin_addon['audio.output'] = AudioOutput(self.window)

        # schedule
        self.window.ui.plugin_addon['schedule'] = ChatStatusLabel("")

        # inline vision
        self.window.ui.nodes['inline.vision'] = HelpLabel(trans('inline.vision'))
        self.window.ui.nodes['inline.vision'].setVisible(False)
        self.window.ui.nodes['inline.vision'].setContentsMargins(0, 0, 0, 0)

        opts_layout = QHBoxLayout()
        # opts_layout.setSpacing(2)  #
        opts_layout.setContentsMargins(0, 0, 0, 0)
        opts_layout.addWidget(self.window.ui.nodes['output.timestamp'])
        # opts_layout.addWidget(self.window.ui.nodes['output.edit'])
        opts_layout.addWidget(self.window.ui.nodes['output.raw'])
        opts_layout.addWidget(self.window.ui.nodes['inline.vision'])
        opts_layout.setAlignment(Qt.AlignLeft)

        left_layout = QHBoxLayout()
        left_layout.addLayout(opts_layout)
        left_layout.setContentsMargins(0, 0, 0, 0)

        right_layout = QHBoxLayout()
        right_layout.addWidget(self.window.ui.nodes['icon.video.capture'])
        right_layout.addWidget(self.window.ui.nodes['icon.audio.input'])
        right_layout.addWidget(self.window.ui.nodes['icon.audio.output'])
        right_layout.addWidget(self.window.ui.nodes['icon.interpreter'])
        right_layout.addWidget(self.window.ui.nodes['icon.indexer'])
        right_layout.addWidget(self.window.ui.plugin_addon['schedule'])
        right_layout.addWidget(QLabel(" "))
        right_layout.addWidget(self.window.ui.nodes['chat.plugins'])
        right_layout.addWidget(QLabel(" "))
        right_layout.addWidget(self.window.ui.nodes['chat.label'])
        right_layout.addWidget(QLabel("  "))
        right_layout.addWidget(self.window.ui.nodes['chat.model'])
        right_layout.addWidget(QLabel("  "))
        right_layout.addWidget(self.window.ui.nodes['prompt.context'])
        right_layout.addWidget(QLabel("  "))
        right_layout.addWidget(self.window.ui.nodes['input.counter'])
        right_layout.setContentsMargins(0, 8, 20, 0)

        left_widget = QWidget()
        left_widget.setLayout(left_layout)
        right_widget = QWidget()
        right_widget.setLayout(right_layout)

        self.window.ui.nodes['anim.loading'] = Loading()
        self.window.ui.nodes['anim.loading'].hide()

        grid = QGridLayout()

        left_layout = QHBoxLayout()
        left_layout.addWidget(left_widget)
        #left_layout.addStretch(1)
        left_layout.setContentsMargins(0, 0, 0, 0)

        center_layout = QHBoxLayout()
        #center_layout.addStretch()
        center_layout.addWidget(self.window.ui.nodes['anim.loading'])
        center_layout.addStretch()
        center_layout.setContentsMargins(20, 0, 0, 0)

        right_layout = QHBoxLayout()
        right_layout.addStretch(1)
        right_layout.addWidget(right_widget)
        right_layout.setContentsMargins(0, 0, 0, 0)

        grid.addLayout(left_layout, 0, 0)
        grid.addLayout(center_layout, 0, 1, alignment=Qt.AlignCenter)
        grid.addLayout(right_layout, 0, 2, alignment=Qt.AlignRight)
        grid.setContentsMargins(0, 0, 0, 0)

        self.window.ui.nodes['chat.footer'] = QWidget()
        self.window.ui.nodes['chat.footer'].setLayout(grid)

        bottom_layout = QVBoxLayout()
        bottom_layout.addWidget(self.window.ui.nodes['chat.footer'])
        bottom_layout.setContentsMargins(2, 0, 2, 0)
        return bottom_layout

    def set_fixed_size_policy(self, layout):
        for i in range(layout.count()):
            item = layout.itemAt(i)
            if isinstance(item, QWidgetItem):
                widget = item.widget()
                if widget:
                    widget.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
            elif isinstance(item, QLayout):
                self.set_fixed_size_policy(item)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

import os
from typing import Optional, List, Dict

from llama_index.core.base.embeddings.base import BaseEmbedding
from llama_index.core.llms.llm import BaseLLM as LlamaBaseLLM
from llama_index.core.multi_modal_llms import MultiModalLLM as LlamaMultiModalLLM

from pygpt_net.core.types import (
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
)
from pygpt_net.item.model import ModelItem
from pygpt_net.utils import parse_args


class BaseLLM:
    def __init__(self, *args, **kwargs):
        self.id = ""
        self.name = ""
        self.type = []  # langchain, llama_index, embeddings
        self.description = ""

    def init(
            self,
            window,
            model: ModelItem,
            mode: str,
            sub_mode: str = None
    ):
        """
        Initialize provider

        :param window: window instance
        :param model: model instance
        :param mode: mode (langchain, llama_index)
        :param sub_mode: sub mode (chat, completion)
        """
        options = {}
        if mode == MODE_LANGCHAIN:
            options = model.langchain
        elif mode == MODE_LLAMA_INDEX:
            options = model.llama_index
        if 'env' in options:
            for item in options['env']:
                if item['name'] is None or item['name'] == "":
                    continue
                try:
                    os.environ[item['name']] = str(item['value'].format(**window.core.config.all()))
                except Exception as e:
                    pass

    def init_embeddings(
            self,
            window,
            env: Optional[List[Dict]] = None
    ):
        """
        Initialize embeddings provider

        :param window: window instance
        :param env: ENV configuration list
        """
        if env is not None and len(env) > 0:
            for item in env:
                if item['name'] is None or item['name'] == "":
                    continue
                try:
                    os.environ[item['name']] = str(item['value'].format(**window.core.config.all()))
                except Exception as e:
                    pass

    def parse_args(
            self,
            options: dict
    ) -> dict:
        """
        Parse extra args

        :param options: LLM options dict (langchain, llama_index)
        :return: parsed arguments dict
        """
        args = {}
        if 'args' in options:
            if options['args'] is not None:
                args = parse_args(options['args'])
        return args

    def completion(
            self,
            window,
            model: ModelItem,
            stream: bool = False
    ) -> any:
        """
        Return LLM provider instance for completion in langchain mode

        :param window: window instance
        :param model: model instance
        :param stream: stream mode
        :return: provider instance
        """
        pass

    def chat(
            self,
            window,
            model: ModelItem,
            stream: bool = False
    ) -> any:
        """
        Return LLM provider instance for chat in langchain mode

        :param window: window instance
        :param model: model instance
        :param stream: stream mode
        :return: provider instance
        """
        pass

    def llama(
            self,
            window,
            model: ModelItem,
            stream: bool = False
    ) -> LlamaBaseLLM:
        """
        Return LLM provider instance for llama index query and chat

        :param window: window instance
        :param model: model instance
        :param stream: stream mode
        :return: provider instance
        """
        pass

    def llama_multimodal(
            self,
            window,
            model: ModelItem,
            stream: bool = False
    ) -> LlamaMultiModalLLM:
        """
        Return multimodal LLM provider instance for llama

        :param window: window instance
        :param model: model instance
        :param stream: stream mode
        :return: LLM provider instance
        """
        pass

    def get_embeddings_model(
            self,
            window,
            config: Optional[List[Dict]] = None
    ) -> BaseEmbedding:
        """
        Return provider instance for embeddings

        :param window: window instance
        :param config: config keyword arguments list
        :return: provider instance
        """
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Dict, Any


class ReplyContext:

    AGENT_CONTINUE = "agent.continue"
    CMD_EXECUTE = "cmd.execute"
    CMD_EXECUTE_INLINE = "cmd.execute.inline"
    EXPERT_CALL = "expert.call"
    EXPERT_RESPONSE = "expert.response"

    def __init__(self):
        """Reply context"""
        self.type = None
        self.bridge_context = None
        self.ctx = None
        self.prev_ctx = None
        self.parent_id = None
        self.input = ""
        self.cmds = []

    def to_dict(self) -> Dict[str, Any]:
        """
        Dump to dictionary

        :return: dict
        """
        data = {
            "type": self.type,
            "bridge_context": self.bridge_context,
            "ctx": self.ctx,
            "prev_ctx": self.prev_ctx,
            "parent_id": self.parent_id,
            "input": self.input,
            "cmds": self.cmds,
        }
        if self.bridge_context is not None:
            data["bridge_context"] = self.bridge_context.to_dict()
        if self.ctx is not None:
            data["ctx"] = self.ctx.to_dict()
        if self.prev_ctx is not None:
            data["prev_ctx"] = self.prev_ctx.to_dict()
        return data

class Reply:
    def __init__(self, window=None):
        """
        Reply core

        :param window: Window instance
        """
        self.window = window


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.19 02:00:00                  #
# ================================================== #

import uuid
from datetime import datetime
from typing import Optional, Any, Dict

from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QVBoxLayout, QWidget, QLayout

from pygpt_net.ui.widget.tabs.body import TabBody
from pygpt_net.utils import trans

from .tab import Tab


class Tabs:

    # number of columns
    NUM_COLS = 2

    def __init__(self, window=None):
        """
        Tabs core

        :param window: Window instance
        """
        self.window = window
        self.last_pid = -1
        self.pids = {}  # pid: Tab data
        self.icons = {
            Tab.TAB_CHAT: ":/icons/chat.svg",
            Tab.TAB_NOTEPAD: ":/icons/paste.svg",
            Tab.TAB_FILES: ":/icons/folder_filled.svg",
            Tab.TAB_TOOL_PAINTER: ":/icons/brush.svg",
            Tab.TAB_TOOL_CALENDAR: ":/icons/calendar.svg",
            Tab.TAB_TOOL: ":/icons/build.svg",
        }
        self.titles = {
            Tab.TAB_CHAT: "output.tab.chat",
            Tab.TAB_NOTEPAD: "output.tab.notepad",
            Tab.TAB_FILES: "output.tab.files",
            Tab.TAB_TOOL_PAINTER: "output.tab.painter",
            Tab.TAB_TOOL_CALENDAR: "output.tab.calendar",
            Tab.TAB_TOOL: "output.tab.tool",
        }

    def get_tab_by_index(
            self,
            idx: int,
            column_idx: int = 0
    ) -> Optional[Tab]:
        """
        Get tab by index

        :param idx: Tab index
        :param column_idx: Column index
        :return: Tab
        """
        tab = self.window.ui.layout.get_tabs_by_idx(column_idx).widget(idx)
        if tab is None:
            return None
        return tab.getOwner()

    def get_tab_by_pid(self, pid: int) -> Optional[Tab]:
        """
        Get tab by PID

        :param pid: PID
        :return: Tab
        """
        if pid in self.pids:
            return self.pids[pid]
        return None

    def get_first_tab_by_type(self, type: int) -> Optional[Tab]:
        """
        Get first tab by type

        :param type: Tab type
        :return: Tab
        """
        for pid in self.pids:
            tab = self.pids[pid]
            if tab.type == type:
                return tab
        return None

    def get_active_pid(self) -> Optional[int]:
        """
        Get PID by active tab

        :return: PID
        """
        current_column_idx = self.window.controller.ui.tabs.get_current_column_idx()
        tabs = self.window.ui.layout.get_tabs_by_idx(current_column_idx)
        tab = self.get_tab_by_index(tabs.currentIndex(), current_column_idx)
        if tab is None:
            return 0
        return tab.pid

    def add(
            self,
            type: int,
            title: str,
            icon: Optional[str] = None,
            child: Optional[TabBody] = None,
            data_id: Optional[int] = None,
            tool_id: Optional[str] = None
    ) -> Tab:
        """
        Add tab

        :param type: Tab type
        :param title: Tab title
        :param icon: Tab icon
        :param child: Tab child (TabBody)
        :param data_id: Tab data ID
        :param tool_id: Tool ID
        :return: Tab
        """
        self.last_pid += 1  # PID++, start from 0

        tab = Tab()
        tab.uuid = uuid.uuid4()
        tab.pid = self.last_pid
        tab.type = type
        tab.title = title
        tab.icon = icon
        tab.child = child
        tab.data_id = data_id
        tab.tool_id = tool_id

        if type == Tab.TAB_CHAT:
            self.add_chat(tab)
        elif type == Tab.TAB_NOTEPAD:
            self.add_notepad(tab)
        elif type == Tab.TAB_FILES:
            self.add_tool_explorer(tab)
        elif type == Tab.TAB_TOOL_PAINTER:
            self.add_tool_painter(tab)
        elif type == Tab.TAB_TOOL_CALENDAR:
            self.add_tool_calendar(tab)
        elif type == Tab.TAB_TOOL:
            self.add_tool(tab)

        self.pids[tab.pid] = tab
        return tab

    def append(
            self,
            type: int,
            tool_id: str,
            idx: int,
            column_idx: int = 0
    ) -> Tab:
        """
        Append tab to the right side of the tab with the specified index

        :param type: tab type
        :param tool_id: tool ID
        :param idx: index of the tab to the right of which the new tab will be added
        :param column_idx: index of the column in which the tab will be added
        :return: Tab
        """
        self.last_pid += 1  # PID++, start from 0
        title = ""
        icon = self.icons[type]
        if type == Tab.TAB_CHAT:
            title = trans('output.tab.chat') + " {}".format(self.count_by_type(type) + 1)
        elif type == Tab.TAB_NOTEPAD:
            title = trans('output.tab.notepad') + " {}".format(self.count_by_type(type) + 1)

        tab = Tab()
        tab.uuid = uuid.uuid4()
        tab.pid = self.last_pid
        tab.type = type
        tab.title = title
        tab.icon = icon
        tab.new_idx = idx + 1  # place on right side
        tab.column_idx = column_idx
        tab.tool_id = tool_id

        if type == Tab.TAB_CHAT:
            self.add_chat(tab)
        elif type == Tab.TAB_NOTEPAD:
            self.add_notepad(tab)
        elif type == Tab.TAB_TOOL:
            self.add_tool(tab)

        self.pids[tab.pid] = tab
        self.update()

        # load data from db
        if type == Tab.TAB_NOTEPAD:
            self.window.controller.notepad.load()
        return tab

    def restore(self, data: dict):
        """
        Restore tab from saved data

        :param data: Tab data
        """
        tab = Tab()
        tab.uuid = data["uuid"]
        tab.pid = data["pid"]
        tab.type = data["type"]
        tab.title = data["title"]
        tab.data_id = data["data_id"]
        tab.child = None

        if 'tooltip' in data and data['tooltip'] is not None:
            tab.tooltip = data['tooltip']

        if 'custom_name' in data and data['custom_name'] is not None:
            tab.custom_name = data['custom_name']

        if 'column_idx' in data and data['column_idx'] is not None:
            tab.column_idx = data['column_idx']

        if 'tool_id' in data and data['tool_id'] is not None:
            tab.tool_id = data['tool_id']

        if tab.type in self.icons:
            tab.icon = self.icons[tab.type]

        if tab.type == Tab.TAB_CHAT:  # chat
            self.add_chat(tab)
            self.window.core.ctx.output.mapping[tab.pid] = tab.data_id  # restore pid => meta.id mapping
            self.window.core.ctx.output.last_pids[tab.data_id] = tab.pid
            self.window.core.ctx.output.last_pid = tab.pid
        elif tab.type == Tab.TAB_NOTEPAD: # notepad
            self.add_notepad(tab)
        elif tab.type == Tab.TAB_FILES:  # files
            self.add_tool_explorer(tab)
        elif tab.type == Tab.TAB_TOOL_PAINTER:  # painter
            self.add_tool_painter(tab)
        elif tab.type == Tab.TAB_TOOL_CALENDAR:  # calendar
            self.add_tool_calendar(tab)
        elif tab.type == Tab.TAB_TOOL:  # custom tools, id 100+
            self.add_tool(tab)

        self.pids[tab.pid] = tab
        self.last_pid = self.get_max_pid()

    def remove_tab_by_idx(
            self,
            idx: int,
            column_idx: int = 0
    ):
        """
        Remove tab by index

        :param idx: Tab index
        :param column_idx: Column index
        """
        tab = self.get_tab_by_index(idx, column_idx)
        if tab is None:
            return
        self.remove(tab.pid)

    def remove(self, pid: int):
        """
        Remove tab by PID

        :param pid: PID
        """
        tab = self.get_tab_by_pid(pid)
        if tab is None:
            return
        column_idx = tab.column_idx
        self.window.ui.layout.get_tabs_by_idx(column_idx).removeTab(tab.idx)
        del self.pids[pid]
        self.update()

    def remove_all(self):
        """Remove all tabs"""
        for pid in list(self.pids):
            self.remove(pid)  # delete from PIDs and UI
        self.window.core.ctx.output.clear()  # clear mapping

    def remove_all_by_type(
            self,
            type: int,
            column_idx: int = 0
    ):
        """
        Remove all tabs by type

        :param type: Tab type
        :param column_idx: Column index
        """
        for pid in list(self.pids):
            tab = self.pids[pid]
            if tab.type == type and tab.column_idx == column_idx:
                if type == Tab.TAB_CHAT:
                    if self.count_by_type(type) == 1:
                        continue  # do not remove last chat tab
                self.remove(pid)

    def reload(self):
        """Reload tabs"""
        self.remove_all()
        self.load()

    def get_tabs_by_type(self, type: int) -> list:
        """
        Get tabs by type

        :param type: Tab type
        :return: List of tabs
        """
        tabs = []
        for pid in self.pids:
            tab = self.pids[pid]
            if tab.type == type:
                tabs.append(tab)
        return tabs

    def get_min_pid(self) -> int:
        """
        Get min PID

        :return: Min PID
        """
        min = 999999
        for pid in self.pids:
            if pid < min:
                min = pid
        return min

    def get_max_pid(self) -> int:
        """
        Get max PID

        :return: Max PID
        """
        max = 0
        for pid in self.pids:
            if pid > max:
                max = pid
        return max

    def count_by_type(self, type: int) -> int:
        """
        Count tabs by type

        :param type: Tab type
        :return: Number of tabs
        """
        count = 0
        for pid in self.pids:
            tab = self.pids[pid]
            if tab.type == type:
                count += 1
        return count

    def get_max_data_id_by_type(self, type: int) -> int:
        """
        Get max data ID by type

        :param type: Tab type
        :return: Max data ID
        """
        max = 0
        for pid in self.pids:
            tab = self.pids[pid]
            if tab.type == type and tab.data_id > max:
                max = tab.data_id
        return max

    def get_order_by_idx_and_type(
            self,
            idx: int,
            type: int
    ) -> int:
        """
        Get the order of the tab by index and type

        :param idx: Tab index
        :param type: Tab type
        :return: Order of the tab (starting from 1)
        """
        tabs_of_type = [tab for tab in self.pids.values() if tab.type == type]
        tabs_of_type.sort(key=lambda tab: tab.idx)

        for order, tab in enumerate(tabs_of_type, start=1):
            if tab.idx == idx:
                return order
        return -1  # Return -1 if the tab with the specified index and type is not found

    def get_min_idx_by_type(
            self,
            type: int,
            column_idx: int = 0
    ) -> int:
        """
        Get min index by type

        :param type: Tab type
        :param column_idx: Column index
        :return: Min index
        """
        min = 999999
        for pid in self.pids:
            tab = self.pids[pid]
            if (tab.type == type
                    and tab.column_idx == column_idx
                    and tab.idx < min):
                min = tab.idx
        return min

    def update(self):
        """Update tabs data (pids) from UI (all columns)"""
        for n in range(0, self.NUM_COLS):
            self.update_column(n)

    def update_column(self, column_idx: int):
        """
        Update column data by index

        :param column_idx: Column index
        """
        tabs = self.window.ui.layout.get_tabs_by_idx(column_idx)
        for pid in self.pids:
            tab = self.pids[pid]
            if tab.column_idx == column_idx:
                tab.idx = tabs.indexOf(tab.child)
                tab.title = tabs.tabText(tab.idx)
                tab.tooltip = tabs.tabToolTip(tab.idx)
                tab.updated_at = datetime.now()

    def add_chat(self, tab: Tab):
        """
        Add chat tab

        :param tab: Tab instance
        """
        column = self.window.ui.layout.get_column_by_idx(tab.column_idx)
        tabs = column.get_tabs()
        tab.parent = column
        tab.child = self.window.core.ctx.container.get(tab)  # tab is already appended here
        if tab.new_idx is not None:
            tab.idx = tabs.insertTab(tab.new_idx, tab.child, tab.title)
        else:
            tab.idx = tabs.addTab(tab.child, tab.title)
        tab.child.setOwner(tab)
        tabs.setTabIcon(tab.idx, QIcon(tab.icon))
        if tab.tooltip is not None:
            tabs.setTabToolTip(tab.idx, tab.tooltip)

    def add_notepad(self, tab: Tab):
        """
        Add notepad tab

        :param tab: Tab instance
        """
        idx = None
        column = self.window.ui.layout.get_column_by_idx(tab.column_idx)
        tabs = column.get_tabs()
        tab.parent = column
        tab.parent = tabs.get_column()
        if tab.data_id is not None:
            idx = tab.data_id  # restore prev idx
        tab.child, idx, data_id = self.window.controller.notepad.create(idx)
        tab.data_id = data_id  # notepad idx in db, enumerated from 1
        if tab.new_idx is not None:
            tab.idx = tabs.insertTab(tab.new_idx, tab.child, tab.title)
        else:
            tab.idx = tabs.addTab(tab.child, tab.title)
        tab.child.setOwner(tab)
        tabs.setTabIcon(tab.idx, QIcon(tab.icon))
        if tab.tooltip is not None:
            tabs.setTabToolTip(tab.idx, tab.tooltip)

    def add_tool_explorer(self, tab: Tab):
        """
        Add explorer tab

        :param tab: Tab instance
        """
        column = self.window.ui.layout.get_column_by_idx(tab.column_idx)
        tabs = column.get_tabs()
        tab.parent = column
        tab.child = self.window.ui.chat.output.explorer.setup()
        tab.idx = tabs.addTab(tab.child, tab.title)
        tab.child.setOwner(tab)
        tabs.setTabIcon(tab.idx, QIcon(tab.icon))
        if tab.tooltip is not None:
            tabs.setTabToolTip(tab.idx, tab.tooltip)

    def add_tool_painter(self, tab: Tab):
        """
        Add painter tab

        :param tab: Tab instance
        """
        column = self.window.ui.layout.get_column_by_idx(tab.column_idx)
        tabs = column.get_tabs()
        tab.parent = column
        tab.child = self.window.ui.chat.output.painter.setup()
        tab.child.append(self.window.ui.painter)
        tab.idx = tabs.addTab(tab.child, tab.title)
        tab.child.setOwner(tab)
        tabs.setTabIcon(tab.idx, QIcon(tab.icon))
        if tab.tooltip is not None:
            tabs.setTabToolTip(tab.idx, tab.tooltip)

    def add_tool_calendar(self, tab: Tab):
        """
        Add calendar tab

        :param tab: Tab instance
        """
        column = self.window.ui.layout.get_column_by_idx(tab.column_idx)
        tabs = column.get_tabs()
        tab.parent = column
        tab.child = self.window.ui.chat.output.calendar.setup()
        tab.idx = tabs.addTab(tab.child, tab.title)
        tab.child.setOwner(tab)
        tabs.setTabIcon(tab.idx, QIcon(tab.icon))
        if tab.tooltip is not None:
            tabs.setTabToolTip(tab.idx, tab.tooltip)

    def add_tool(self, tab: Tab):
        """
        Add custom tool tab

        :param tab: Tab instance
        """
        column = self.window.ui.layout.get_column_by_idx(tab.column_idx)
        tabs = column.get_tabs()
        tool = self.window.tools.get(tab.tool_id)
        if tool is None:
            raise Exception("Tool not found: {}".format(tab.tool_id))
        widget = tool.as_tab(tab)
        if widget is None:
            raise Exception("Tool widget not found: {}".format(tab.tool_id))
        tab.icon = tool.tab_icon
        tab.title = trans(tool.tab_title)
        tab.parent = column
        tab.child = self.from_widget(widget)
        tab.idx = tabs.addTab(tab.child, tab.title)
        tab.child.setOwner(tab)
        tabs.setTabIcon(tab.idx, QIcon(tab.icon))
        if tab.tooltip is not None:
            tabs.setTabToolTip(tab.idx, tab.tooltip)

    def move_tab(self, tab: Tab, column_idx: int):
        """
        Move tab to column

        :param tab: Tab instance
        :param column_idx: Column index
        """
        if tab is None:
            return
        if tab.column_idx == column_idx:
            return

        old_column = self.window.ui.layout.get_column_by_idx(tab.column_idx)
        old_tabs = old_column.get_tabs()
        old_tabs.removeTab(tab.idx)
        new_column = self.window.ui.layout.get_column_by_idx(column_idx)
        new_tabs = new_column.get_tabs()
        tab.idx = new_tabs.addTab(tab.child, QIcon(tab.icon), tab.title)
        tab.parent = new_column
        tab.column_idx = column_idx
        self.update()

    def get_first_by_type(self, type: int) -> Optional[Tab]:
        """
        Get first tab by type

        :param type: Tab type
        :return: Tab
        """
        for pid in self.pids:
            tab = self.pids[pid]
            if tab.type == type:
                return tab
        return None

    def from_defaults(self) -> Dict[int, Any]:
        """
        Prepare default tabs data

        :return: Dict with tabs data
        """
        data = {}
        data[0] = {
            "uuid": uuid.uuid4(),
            "pid": 0,
            "idx": 0,
            "type": Tab.TAB_CHAT,
            "data_id": None,
            "title": "Chat",
            "tooltip": "Chat",
            "column_idx": 0,
        }
        data[1] = {
            "uuid": uuid.uuid4(),
            "pid": 1,
            "idx": 1,
            "type": Tab.TAB_FILES,
            "data_id": None,
            "title": "Files",
            "tooltip": "Files",
            "column_idx": 0,
            "tool_id": "explorer",
        }
        data[2] = {
            "uuid": uuid.uuid4(),
            "pid": 2,
            "idx": 2,
            "type": Tab.TAB_TOOL_CALENDAR,
            "data_id": None,
            "title": "Calendar",
            "tooltip": "Calendar",
            "column_idx": 0,
            "tool_id": "calendar",
        }
        data[3] = {
            "uuid": uuid.uuid4(),
            "pid": 3,
            "idx": 3,
            "type": Tab.TAB_TOOL_PAINTER,
            "data_id": None,
            "title": "Painter",
            "tooltip": "Painter",
            "column_idx": 0,
            "tool_id": "painter",
        }
        """
        data[4] = {
            "uuid": uuid.uuid4(),
            "pid": 4,
            "idx": 4,
            "type": self.TAB_NOTEPAD,
            "data_id": 1,
            "title": "Notepad",
            "tooltip": "Notepad",
            "column_idx": 0,
            "tool_id": "notepad",
        }
        """
        # load notepads from db
        next_idx = 4
        notepads_dict = self.window.core.notepad.import_from_db()
        if notepads_dict is not None:
            for idx in notepads_dict:
                item = notepads_dict[idx]
                data[next_idx] = {
                    "uuid": uuid.uuid4(),
                    "pid": next_idx,
                    "idx": next_idx,
                    "type": Tab.TAB_NOTEPAD,
                    "data_id": item['data_id'],
                    "title": item['title'],
                    "tooltip": item['title'],
                    "column_idx": 0,
                    "tool_id": "notepad",
                }
                next_idx += 1
        return data

    def load(self):
        """Load tabs data from config"""
        data = self.window.core.config.get("tabs.data")
        if data is None or len(data) == 0:
            data = self.from_defaults()

        # check for required tabs
        tmp_pid = -1  # tmp PID only for loading
        required = [Tab.TAB_CHAT, Tab.TAB_FILES, Tab.TAB_TOOL_CALENDAR, Tab.TAB_TOOL_PAINTER]
        for type in required:
            found = False
            for pid in data:
                if data[pid]["type"] == type:
                    found = True
                    break
            if not found:
                data[tmp_pid + 1] = {
                    "uuid": str(uuid.uuid4()),
                    "pid": tmp_pid + 1,
                    "idx": tmp_pid + 1,
                    "type": type,
                    "data_id": None,
                    "title": trans(self.titles[type]),
                    "tooltip": trans(self.titles[type]),
                    "custom_name": False,
                    "column_idx": 0,
                    "tool_id": None,
                }
                tmp_pid += 1

        # sort PIDs by idx
        data = dict(sorted(data.items(), key=lambda item: item[1]["idx"]))
        # regenerate PIDS, enumerate from 0
        n = 0
        for pid in data:
            tab = data[pid]
            tab['pid'] = n
            self.restore(tab)
            n += 1
        self.update()

    def save(self):
        """Save tabs data to config"""
        data = {}
        for pid in self.pids:
            tab = self.pids[pid]
            data[pid] = {
                "uuid": str(tab.uuid),
                "pid": tab.pid,
                "idx": tab.idx,
                "type": tab.type,
                "data_id": tab.data_id,
                "title": tab.title,
                "tooltip": tab.tooltip,
                "custom_name": tab.custom_name,
                "column_idx": tab.column_idx,
                "tool_id": tab.tool_id,
            }
        opened_tabs = {}
        for column_idx in range(0, self.NUM_COLS):
            tabs = self.window.ui.layout.get_tabs_by_idx(column_idx)
            opened_tabs[column_idx] = tabs.currentIndex()
        self.window.core.config.set("tabs.opened", opened_tabs)
        self.window.core.config.set("tabs.data", data)
        self.window.core.config.save()

    def update_title(
            self,
            idx: int,
            title: str,
            tooltip: Optional[str] = None
    ):
        """
        Update tab title

        :param idx: Tab index
        :param title: Tab title
        :param tooltip: Tab tooltip
        """
        column_idx = self.window.controller.ui.tabs.get_current_column_idx()
        tabs = self.window.ui.layout.get_tabs_by_idx(column_idx)
        tab = self.get_tab_by_index(idx, column_idx)
        if tab is None:
            return
        tab.title = title
        tab.tooltip = tooltip
        tab.custom_name = True
        if title is not None:
            tabs.setTabText(idx, title)
        if tooltip is not None:
            tabs.setTabToolTip(idx, tooltip)

    def reload_titles(self):
        """Reload default tab titles"""
        return
        processed = []
        for column_idx in range(0, self.NUM_COLS):
            tabs = self.window.ui.layout.get_tabs_by_idx(column_idx)
            counters = {
                Tab.TAB_CHAT: 1,
                Tab.TAB_NOTEPAD: 1,
                Tab.TAB_FILES: 1,
                Tab.TAB_TOOL_PAINTER: 1,
                Tab.TAB_TOOL_CALENDAR: 1,
            }
            for pid in self.pids:
                tab = self.pids[pid]
                if tab.pid in processed:
                    continue
                if tab.custom_name or tab.type == Tab.TAB_TOOL:
                    continue  # leave custom names and tools
                tab.title = trans(self.titles[tab.type])
                num_tabs = self.count_by_type(tab.type)
                if num_tabs > 1:
                    if tab.type in counters:
                        tab.title += " {}".format(counters[tab.type])
                    else:
                        counters[tab.type] = 1
                    counters[tab.type] += 1
                tabs.setTabText(tab.idx, tab.title)
                tabs.setTabToolTip(tab.idx, tab.title)
                processed.append(pid)

    def from_widget(self, widget: QWidget) -> TabBody:
        """
        Prepare tab body from widget

        :param widget: QWidget
        :return: TabBody
        """
        layout = QVBoxLayout()
        layout.addWidget(widget)
        layout.setContentsMargins(0, 0, 0, 0)
        tab_widget = TabBody(self.window)
        tab_widget.append(widget)
        tab_widget.setLayout(layout)
        return tab_widget

    def from_layout(self, layout: QLayout) -> TabBody:
        """
        Prepare tab body from layout

        :param layout: QLayout
        :return: TabBody
        """
        layout.setContentsMargins(0, 0, 0, 0)
        widget = TabBody(self.window)
        widget.setLayout(layout)
        return widget


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import time
from typing import Any, Dict, Optional, Union, List

from PySide6.QtCore import QObject, Slot

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_AGENT_LLAMA,
    MODE_ASSISTANT,
    MODE_EXPERT,
)
from pygpt_net.core.events import KernelEvent
from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans

from .reply import Reply
from .stack import Stack


class Kernel(QObject):

    STATE_IDLE = "idle"
    STATE_BUSY = "busy"
    STATE_ERROR = "error"

    def __init__(self, window=None):
        super(Kernel, self).__init__(window)
        self.window = window
        self.replies = Reply(window)
        self.stack = Stack(window)
        self.halt = False
        self.busy = False
        self.last_stack = []
        self.status = ""
        self.state = self.STATE_IDLE

    def init(self):
        """Init kernel"""
        self.last_stack = []
        self.halt = False
        self.busy = False
        self.state = self.STATE_IDLE
        self.status = ""

    @Slot(object)
    def listener(self, event: KernelEvent):
        """
        Async listener for kernel events

        :param event: kernel event
        """
        if self.stopped() and event.name != KernelEvent.INPUT_USER:
            return

        self.window.dispatch(event)  # return event to handle()

    def handle(self, event: KernelEvent):
        """
        Handle kernel events

        :param event: kernel event
        """
        if self.stopped() and event.name != KernelEvent.INPUT_USER:
            return

        self.store(event)  # store event in stack
        name = event.name
        context = event.data.get("context")
        extra = event.data.get("extra")
        response = event.data.get("response")

        if name in [
            KernelEvent.INPUT_USER,
            KernelEvent.INPUT_SYSTEM,
        ]:
            response = self.input(context, extra, event)
        elif name in [
            KernelEvent.REQUEST,
            KernelEvent.REQUEST_NEXT,
            KernelEvent.RESPONSE_OK,
            KernelEvent.RESPONSE_ERROR,
            KernelEvent.RESPONSE_FAILED,
            KernelEvent.APPEND_BEGIN,
            KernelEvent.APPEND_DATA,
            KernelEvent.APPEND_END,
            KernelEvent.TOOL_CALL,
            KernelEvent.AGENT_CONTINUE,
            KernelEvent.AGENT_CALL,
            KernelEvent.CALL,
        ]:
            response = self.queue(context, extra, event)
        elif name in [
            KernelEvent.REPLY_ADD,
            KernelEvent.REPLY_RETURN,
        ]:
            response = self.reply(context, extra, event)

        elif name in [
            KernelEvent.STATE_BUSY,
            KernelEvent.STATE_IDLE,
            KernelEvent.STATE_ERROR,
        ]:
            self.set_state(event)

        elif name == KernelEvent.STATUS:
            self.set_status(event.data.get("status"))

        event.data["response"] = response  # update response

    def input(
            self,
            context: BridgeContext,
            extra: Dict[str, Any],
            event: KernelEvent
    ):
        """
        Input message to kernel

        :param context: bridge context
        :param extra: extra data
        :param event: kernel event
        """
        if self.stopped() and event.name != KernelEvent.INPUT_USER:
            return

        if event.name in [
            KernelEvent.INPUT_USER,
            KernelEvent.INPUT_SYSTEM,
        ]:
            return self.window.controller.chat.input.send(context, extra)

    def queue(
            self,
            context: BridgeContext,
            extra: Dict[str, Any],
            event: KernelEvent
    ) -> Optional[Union[bool, str, List[Dict]]]:
        """
        Queue messages to kernel

        :param context: bridge context
        :param extra: extra data
        :param event: kernel event
        :return: response
        """
        if self.stopped():
            return

        if event.name in [
            KernelEvent.REQUEST,
            KernelEvent.REQUEST_NEXT,
        ]:
            return self.call(context, extra, event)
        elif event.name in [
            KernelEvent.RESPONSE_OK,
            KernelEvent.RESPONSE_ERROR,
            KernelEvent.RESPONSE_FAILED,
            KernelEvent.APPEND_BEGIN,
            KernelEvent.APPEND_DATA,
            KernelEvent.APPEND_END,
        ]:
            return self.output(context, extra, event)
        elif event.name in [
            KernelEvent.TOOL_CALL,
            KernelEvent.AGENT_CONTINUE,
            KernelEvent.AGENT_CALL,
        ]:
            return self.stack.add(context.reply_context)  # to reply stack
        elif event.name == KernelEvent.CALL:
            return self.call(context, extra, event)

    def call(
            self,
            context: BridgeContext,
            extra: Dict[str, Any],
            event: KernelEvent
    ) -> Optional[Union[bool, str]]:
        """
        Execute message

        :param context: bridge context
        :param extra: extra data
        :param event: kernel event
        :return: response
        """
        if self.stopped():
            return

        if event.name == KernelEvent.REQUEST:
            return self.window.core.bridge.request(context, extra)
        elif event.name == KernelEvent.REQUEST_NEXT:
            return self.window.core.bridge.request_next(context, extra)
        elif event.name == KernelEvent.CALL:
            return self.window.core.bridge.call(context, extra)

    def reply(
            self,
            context: BridgeContext,
            extra: Dict[str, Any],
            event: KernelEvent
    ) -> Optional[List[Dict]]:
        """
        Queue reply message

        :param context: bridge context
        :param extra: extra data
        :param event: kernel event
        :return: response
        """
        if self.stopped():
            return

        if event.name == KernelEvent.REPLY_ADD:
            return self.replies.add(context, extra)
        elif event.name == KernelEvent.REPLY_RETURN:
            return self.input(context, extra, KernelEvent(KernelEvent.INPUT_SYSTEM))

    def output(
            self,
            context: BridgeContext,
            extra: Dict[str, Any],
            event: KernelEvent
    ):
        """
        Handle output from kernel

        :param context: bridge context
        :param extra: extra data
        :param event: kernel event
        """
        if self.stopped():
            return

        if event.name == KernelEvent.RESPONSE_OK:
            return self.window.controller.chat.response.handle(context, extra, True)
        elif event.name == KernelEvent.RESPONSE_ERROR:
            return self.window.controller.chat.response.handle(context, extra, False)
        elif event.name == KernelEvent.RESPONSE_FAILED:
            return self.window.controller.chat.response.failed(context, extra)
        elif event.name == KernelEvent.APPEND_BEGIN:
            return self.window.controller.chat.response.begin(context, extra)
        elif event.name == KernelEvent.APPEND_DATA:
            return self.window.controller.chat.response.append(context, extra)
        elif event.name == KernelEvent.APPEND_END:
            return self.window.controller.chat.response.end(context, extra)

    def restart(self):
        """Restart kernel"""
        self.window.dispatch(KernelEvent(KernelEvent.RESTART))
        self.init()

    def terminate(self):
        """Terminate kernel"""
        self.window.dispatch(KernelEvent(KernelEvent.TERMINATE))
        self.stop(exit=True)
        self.window.ui.hide_loading()
        self.window.controller.plugins.destroy()

    def stop(self, exit: bool = False):
        """
        Stop kernel

        :param exit: on app exit
        """
        self.halt = True
        self.window.controller.chat.common.stop(exit=exit)  # it stops legacy agent also
        self.window.controller.audio.stop_audio()
        if not exit:
            self.window.dispatch(KernelEvent(KernelEvent.STOP))
            self.set_state(KernelEvent(KernelEvent.STATE_IDLE, {"msg": trans("status.stopped")}))

    def set_state(self, event: KernelEvent):
        """
        Set kernel state

        :param: KernelEvent event
        """
        # update state
        if event.name == KernelEvent.STATE_BUSY:
            self.busy = True
            self.state = self.STATE_BUSY
            self.window.ui.tray.set_icon(self.STATE_BUSY)
            if not self.halt:
                self.window.ui.show_loading()
        elif event.name == KernelEvent.STATE_IDLE:
            self.busy = False
            self.state = self.STATE_IDLE
            self.window.ui.tray.set_icon(self.STATE_IDLE)
            self.window.ui.hide_loading()
        elif event.name == KernelEvent.STATE_ERROR:
            self.busy = False
            self.state = self.STATE_ERROR
            self.window.ui.tray.set_icon(self.STATE_ERROR)
            self.window.ui.hide_loading()

        # update message if provided
        msg = event.data.get("msg", None)
        if msg is not None:
            self.set_status(msg)

    def set_status(self, status: str):
        """
        Set kernel status

        :param status: status
        """
        self.status = status
        self.window.ui.status(status)

    def resume(self):
        """Resume kernel"""
        self.halt = False

    def stopped(self) -> bool:
        """
        Check if kernel is stopped

        :return: True if stopped
        """
        return self.halt

    def store(self, event):
        """
        Store event in stack

        :param event: event
        """
        # last 30 events
        if len(self.last_stack) > 30:
            self.last_stack.pop(0)
        ts = time.strftime("%H:%M:%S: ", time.localtime())
        self.last_stack.append(ts + event.name)

    def async_allowed(self, ctx: CtxItem) -> bool:
        """
        Check if async execution are allowed

        :param ctx: context item
        :return: True if async commands are allowed
        """
        disabled = [MODE_ASSISTANT, MODE_AGENT, MODE_EXPERT, MODE_AGENT_LLAMA]
        if self.window.core.config.get("mode") in disabled:
            return False
        if ctx.agent_call:
            return False
        if self.window.controller.agent.legacy.enabled() or self.window.controller.agent.experts.enabled():
            return False
        return True

    def is_threaded(self) -> bool:
        """
        Check if plugin run is threaded

        :return: True if threaded
        """
        if self.window.core.config.get("mode") == MODE_AGENT_LLAMA:
            return True
        return False


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 19:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        """
        Set default options for plugin

        :param plugin: plugin instance
        """
        plugin.add_option(
            "use_tags",
            type="bool",
            value=False,
            label="Enable: using context @ ID tags",
            description="When enabled, it allows to automatically retrieve context history using @ tags, "
                        "e.g. use @123 in question to retrieve summary of context with ID 123",
        )
        plugin.add_option(
            "model_summarize",
            type="combo",
            value="gpt-4o-mini",
            label="Model",
            description="Model used for summarize, default: gpt-3.5-turbo",
            tooltip="Summarize model",
            use="models",
        )
        plugin.add_option(
            "summary_max_tokens",
            type="int",
            value=1500,
            label="Max summary tokens",
            description="Max tokens in output when generating summary",
            min=0,
            max=None,
        )
        plugin.add_option(
            "ctx_items_limit",
            type="int",
            value=30,
            label="Max contexts to retrieve",
            description="Max items in context history list to retrieve in one query. 0 = no limit",
            min=0,
            max=None,
        )
        plugin.add_option(
            "chunk_size",
            type="int",
            value=100000,
            label="Per-context items content chunk size",
            description="Per-context content chunk size (max characters per chunk)",
            min=1,
            max=None,
        )
        plugin.add_option(
            "prompt_tag_system",
            type="textarea",
            value="ADDITIONAL CONTEXT: Use the following JSON summary of previous discussions as additional context, "
                  "instead of using commands for retrieve content: {context}",
            label="Prompt: tag_system",
            description="Prompt for use @ tag (system)",
            advanced=True,
        )
        plugin.add_option(
            "prompt_tag_summary",
            type="textarea",
            value="You are an expert in context summarization. "
                  "Please summarize the given discussion (ID: {id}) by addressing the query and preparing it to serve "
                  "as additional context for a new discussion or to continue the current one: {query}",
            label="Prompt: tag_summary",
            description="Prompt for use @ tag (summary)",
            advanced=True,
        )

        # commands
        plugin.add_cmd(
            "get_ctx_list_in_date_range",
            instruction="get list of context history (previous conversations between you and me), using date-range "
                        "syntax: \"@date(YYYY-MM-DD)\" for single day, \"@date(YYYY-MM-DD,)\" for date FROM, "
                        "\"@date(,YYYY-MM-DD)\" for date TO, and \"@date(YYYY-MM-DD,YYYY-MM-DD)\" for date FROM-TO",
            params=[
                {
                    "name": "range_query",
                    "type": "str",
                    "description": "range query",
                    "required": True,
                },
            ],
            enabled=True,
            description="When enabled, it allows getting the list of context history (previous conversations)",
        )
        plugin.add_cmd(
            "get_ctx_content_by_id",
            instruction="get summarized content of context by its ID, use summary query to ask another "
                        "model to summarize content, e.g. \"Summarize following discussion answering the query: (query)\"",
            params=[
                {
                    "name": "id",
                    "type": "int",
                    "description": "context ID",
                    "required": True,
                },
                {
                    "name": "summary_query",
                    "type": "str",
                    "description": "query",
                    "required": True,
                },
            ],
            enabled=True,
            description="When enabled, it allows getting summarized content of context with defined ID",
        )
        plugin.add_cmd(
            "count_ctx_in_date",
            instruction="count items of context history (previous conversations between us), by providing year, "
                        "month, day or combination of them",
            params=[
                {
                    "name": "year",
                    "type": "int",
                    "description": "year",
                    "required": True,
                },
                {
                    "name": "month",
                    "type": "int",
                    "description": "month",
                    "required": True,
                },
                {
                    "name": "day",
                    "type": "int",
                    "description": "day",
                    "required": True,
                },
            ],
            enabled=True,
            description="When enabled, it allows counting contexts in date range",
        )
        plugin.add_cmd(
            "get_day_note",
            instruction="get day notes for date",
            params=[
                {
                    "name": "year",
                    "type": "int",
                    "description": "year",
                    "required": True,
                },
                {
                    "name": "month",
                    "type": "int",
                    "description": "month",
                    "required": True,
                },
                {
                    "name": "day",
                    "type": "int",
                    "description": "day",
                    "required": True,
                },
            ],
            enabled=True,
            description="When enabled, it allows retrieving day note for specific date",
        )
        plugin.add_cmd(
            "add_day_note",
            instruction="add day note",
            params=[
                {
                    "name": "note",
                    "type": "str",
                    "description": "content",
                    "required": True,
                },
                {
                    "name": "year",
                    "type": "int",
                    "description": "year",
                    "required": True,
                },
                {
                    "name": "month",
                    "type": "int",
                    "description": "month",
                    "required": True,
                },
                {
                    "name": "day",
                    "type": "int",
                    "description": "day",
                    "required": True,
                },
            ],
            enabled=True,
            description="When enabled, it allows adding day note for specific date",
        )
        plugin.add_cmd(
            "update_day_note",
            instruction="update day note",
            params=[
                {
                    "name": "note",
                    "type": "str",
                    "description": "content",
                    "required": True,
                },
                {
                    "name": "year",
                    "type": "int",
                    "description": "year",
                    "required": True,
                },
                {
                    "name": "month",
                    "type": "int",
                    "description": "month",
                    "required": True,
                },
                {
                    "name": "day",
                    "type": "int",
                    "description": "day",
                    "required": True,
                },
            ],
            enabled=True,
            description="When enabled, it allows updating day note for specific date",
        )
        plugin.add_cmd(
            "remove_day_note",
            instruction="remove day note",
            params=[
                {
                    "name": "year",
                    "type": "int",
                    "description": "year",
                    "required": True,
                },
                {
                    "name": "month",
                    "type": "int",
                    "description": "month",
                    "required": True,
                },
                {
                    "name": "day",
                    "type": "int",
                    "description": "day",
                    "required": True,
                },
            ],
            enabled=True,
            description="When enabled, it allows removing day note for specific date",
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.24 22:00:00                  #
# ================================================== #
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QCheckBox, QHBoxLayout, QWidget, QLabel

from pygpt_net.ui.widget.anims.toggles import AnimToggle
from pygpt_net.utils import trans


class ToggleLabel(QWidget):
    def __init__(
            self,
            title: str = None,
            label_position: str = 'right',
            icon=None,
            icon_size=24,
            parent=None,
    ):
        """
        Toggle checkbox with label

        :param title: label title
        """
        super(ToggleLabel, self).__init__()
        self.title = title
        self.label = QLabel(self.title)
        self.box = AnimToggle()

        ico = None
        if icon is not None:
            ico = QLabel()
            pixmap = QIcon(icon).pixmap(icon_size, icon_size)
            ico.setPixmap(pixmap)

        self.layout = QHBoxLayout()
        if label_position == 'left':
            if icon is not None:
                self.layout.addWidget(ico)
            self.layout.addWidget(self.label)
            self.layout.addWidget(self.box)
        else:
            self.layout.addWidget(self.box)
            if icon is not None:
                self.layout.addWidget(ico)
            self.layout.addWidget(self.label)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)

    def setText(self, text: str):
        """
        Set label text

        :param text: text
        """
        self.label.setText(text)

    def setChecked(self, state: bool):
        """
        Set checkbox state

        :param state: state
        """
        self.box.setChecked(state)

    def isChecked(self) -> bool:
        """
        Get checkbox state

        :return: state
        """
        return self.box.isChecked()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.17 03:00:00                  #
# ================================================== #

import os

from PySide6.QtGui import QPixmap, Qt, QIcon
from PySide6.QtWidgets import QVBoxLayout, QLabel, QPushButton, QHBoxLayout, QPlainTextEdit

from pygpt_net.ui.widget.dialog.info import InfoDialog
from pygpt_net.utils import trans


class About:
    def __init__(self, window=None):
        """
        About dialog

        :param window: Window instance
        """
        self.window = window
        self.thanks = None

    def get_thanks(self) -> str:
        """
        Get people list

        :return: contributors and supporters list
        """
        return self.window.core.updater.get_fetch_thanks()

    def prepare_content(self) -> str:
        """
        Get info text

        :return: info text
        """
        llama_index_version = None
        langchain_version = None
        openai_version = None
        versions = True

        try:
            from llama_index.core import __version__ as llama_index_version
            from langchain import __version__ as langchain_version
            from openai.version import VERSION as openai_version
        except ImportError:
            pass

        lib_versions = ""
        if llama_index_version is None or langchain_version is None or openai_version is None:
            versions = False

        if versions:
            lib_versions = "OpenAI API: {}, LangChain: {}, LlamaIndex: {}\n\n".format(
                openai_version,
                langchain_version,
                llama_index_version,
            )

        platform = self.window.core.platforms.get_as_string()
        version = self.window.meta['version']
        build = self.window.meta['build']
        website = self.window.meta['website']
        github = self.window.meta['github']
        docs = self.window.meta['docs']
        author = self.window.meta['author']
        email = self.window.meta['email']

        label_version = trans("dialog.about.version")
        label_build = trans("dialog.about.build")
        label_website = trans("dialog.about.website")
        label_github = trans("dialog.about.github")
        label_docs = trans("dialog.about.docs")

        data = "{label_version}: {version}, {platform}\n" \
               "{label_build}: {build}\n" \
               "{lib_versions}" \
               "{label_website}: {website}\n" \
               "{label_github}: {github}\n" \
               "{label_docs}: {docs}\n\n" \
               "(c) 2025 {author}\n" \
               "{email}\n".format(
                label_version=label_version,
                version=version,
                platform=platform,
                label_build=label_build,
                build=build.replace('.', '-'),
                label_website=label_website,
                website=website,
                label_github=label_github,
                github=github,
                label_docs=label_docs,
                docs=docs,
                author=author,
                email=email,
                lib_versions=lib_versions,
            )
        return data

    def setup(self):
        """Setups about dialog"""
        id = 'about'

        logo_label = QLabel()
        path = os.path.abspath(os.path.join(self.window.core.config.get_app_path(), 'data', 'logo.png'))
        pixmap = QPixmap(path)
        logo_label.setPixmap(pixmap)

        self.window.ui.nodes['dialog.about.btn.website'] = QPushButton(QIcon(":/icons/home.svg"), trans('about.btn.website'))
        self.window.ui.nodes['dialog.about.btn.website'].clicked.connect(lambda: self.window.controller.dialogs.info.goto_website())
        self.window.ui.nodes['dialog.about.btn.website'].setCursor(Qt.PointingHandCursor)
        self.window.ui.nodes['dialog.about.btn.website'].setStyleSheet("font-size: 11px;")

        self.window.ui.nodes['dialog.about.btn.github'] = QPushButton(QIcon(":/icons/code.svg"), trans('about.btn.github'))
        self.window.ui.nodes['dialog.about.btn.github'].clicked.connect(lambda: self.window.controller.dialogs.info.goto_github())
        self.window.ui.nodes['dialog.about.btn.github'].setCursor(Qt.PointingHandCursor)

        self.window.ui.nodes['dialog.about.btn.support'] = QPushButton(QIcon(":/icons/favorite.svg"), trans('about.btn.support'))
        self.window.ui.nodes['dialog.about.btn.support'].clicked.connect(lambda: self.window.controller.dialogs.info.goto_donate())
        self.window.ui.nodes['dialog.about.btn.support'].setCursor(Qt.PointingHandCursor)

        buttons_layout = QHBoxLayout()
        buttons_layout.addWidget(self.window.ui.nodes['dialog.about.btn.support'])
        buttons_layout.addWidget(self.window.ui.nodes['dialog.about.btn.website'])
        buttons_layout.addWidget(self.window.ui.nodes['dialog.about.btn.github'])

        string = self.prepare_content()
        self.window.ui.nodes['dialog.about.content'] = QLabel(string)
        self.window.ui.nodes['dialog.about.content'].setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.window.ui.nodes['dialog.about.thanks'] = QLabel(trans('about.thanks'))
        self.window.ui.nodes['dialog.about.thanks'].setAlignment(Qt.AlignCenter)

        title = QLabel("PyGPT")
        title.setContentsMargins(0, 0, 0, 0)
        title.setStyleSheet(
            "font-size: 16px; "
            "font-weight: bold; "
            "margin-bottom: 10px; "
            "margin-left: 0; "
            "margin-top: 10px; "
            "padding: 0;"
        )
        title.setAlignment(Qt.AlignCenter)

        self.window.ui.nodes['dialog.about.thanks.content'] = QPlainTextEdit()
        self.window.ui.nodes['dialog.about.thanks.content'].setReadOnly(True)
        self.window.ui.nodes['dialog.about.thanks.content'].setPlainText("")
        self.window.ui.nodes['dialog.about.thanks.content'].setStyleSheet("font-size: 11px;")

        layout = QVBoxLayout()
        layout.addWidget(logo_label)
        layout.addWidget(title)
        layout.addWidget(self.window.ui.nodes['dialog.about.content'])
        layout.addWidget(self.window.ui.nodes['dialog.about.thanks'])
        layout.addWidget(self.window.ui.nodes['dialog.about.thanks.content'])
        layout.addStretch(1)
        layout.addLayout(buttons_layout)

        self.window.ui.dialog['info.' + id] = InfoDialog(self.window, id)
        self.window.ui.dialog['info.' + id].setLayout(layout)
        self.window.ui.dialog['info.' + id].setWindowTitle(trans("dialog.about.title"))

    def prepare(self):
        """Update dialog content"""
        people = str(self.get_thanks())
        self.window.ui.nodes['dialog.about.thanks.content'].setPlainText(people)
        if people == "":
            self.window.ui.nodes['dialog.about.thanks'].hide()
            self.window.ui.nodes['dialog.about.thanks.content'].hide()
        else:
            self.window.ui.nodes['dialog.about.thanks'].show()
            self.window.ui.nodes['dialog.about.thanks.content'].show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.16 01:00:00                  #
# ================================================== #

import copy
import datetime
import json
import time


class CtxItem:
    def __init__(self, mode=None):
        """
        Context item

        :param mode: Mode (completion, chat, img, vision, langchain, assistant, llama_index, agent, expert)
        """
        self.id = None
        self.meta = None  # CtxMeta object
        self.meta_id = None
        self.external_id = None
        self.stream = None
        self.cmds = []
        self.cmds_before = []
        self.results = []
        self.urls = []
        self.urls_before = []
        self.images = []
        self.images_before = []
        self.files = []
        self.files_before = []
        self.attachments = []
        self.attachments_before = []
        self.additional_ctx = []
        self.reply = False
        self.input = None
        self.output = None
        self.mode = mode
        self.model = None
        self.thread = None
        self.msg_id = None
        self.run_id = None
        self.input_name = None
        self.output_name = None
        self.input_timestamp = None
        self.output_timestamp = None
        self.hidden_input = None
        self.hidden_output = None
        self.input_tokens = 0
        self.output_tokens = 0
        self.total_tokens = 0
        self.extra = {}
        self.extra_ctx = None
        self.current = False
        self.internal = False
        self.is_vision = False
        self.is_audio = False
        self.idx = 0
        self.first = False
        self.live = False  # True if is current flow (not loaded from DB)
        self.agent_call = False # prevents plugin reply if True
        self.tool_calls = []  # API tool calls
        self.index_meta = {}  # llama-index metadata ctx used
        self.doc_ids = []  # document ids
        self.prev_ctx = None  # previous context (reply output)
        self.stopped = False  # run stopped
        self.sub_calls = 0  # sub calls count
        self.sub_call = False  # is sub call
        self.sub_reply = False  # sub call reply
        self.hidden = False  # hidden context
        self.pid = 0
        self.audio_id = None
        self.audio_output = None
        self.audio_expires_ts = 0


    @property
    def final_input(self) -> str:
        """
        Final input

        :return: input text
        """
        content = None
        if self.input is not None:
            content = self.input
            if self.hidden_input:
                content += "\n\n" + self.hidden_input
        return content

    @property
    def final_output(self) -> str:
        """
        Final output

        :return: output text
        """
        content = None
        if self.output is not None:
            content = self.output
            if self.hidden_output:
                content += "\n\n" + self.hidden_output
        return content

    def clear_reply(self):
        """Clear current reply output"""
        if self.reply:
            self.urls = []
            self.urls_before = []

    def from_previous(self):
        """Copy data from previous context reply to current context"""
        if self.prev_ctx is not None:
            self.urls = copy.deepcopy(self.prev_ctx.urls)
            self.images = copy.deepcopy(self.prev_ctx.images)
            if self.prev_ctx.images_before:
                self.images = copy.deepcopy(self.prev_ctx.images_before)
            if self.prev_ctx.files_before:
                self.files = copy.deepcopy(self.prev_ctx.files_before)
            if self.prev_ctx.attachments_before:
                self.attachments = copy.deepcopy(self.prev_ctx.attachments_before)
            if self.prev_ctx.urls_before:
                self.urls = copy.deepcopy(self.prev_ctx.urls_before)

    def has_commands(self) -> bool:
        """
        Check if commands are

        :return: True if commands are present
        """
        return len(self.cmds) > 0 or len(self.tool_calls) > 0

    def audio_read_allowed(self) -> bool:
        """
        Check if audio read allowed

        :return: True if audio read allowed
        """
        allowed = True
        if self.has_commands() or '~###~{"cmd":' in self.output:
            allowed = False
        return allowed

    def add_doc_meta(self, meta: dict):
        """
        Add document meta

        :param meta: document meta data
        """
        self.index_meta = meta
        self.doc_ids.append(meta)

    def set_input(self, input: str | None, name: str = None):
        """
        Set input

        :param input: input text (prompt)
        :param name: input person name
        """
        self.input = input
        self.input_name = name
        self.input_timestamp = int(time.time())

    def set_output(self, output: str | None, name: str = None):
        """
        Set output

        :param output: output text
        :param name: output person name
        """
        self.output = output
        self.output_name = name
        self.output_timestamp = int(time.time())

    def set_tokens(self, input_tokens: int, output_tokens: int):
        """
        Set tokens usage

        :param input_tokens: prompt tokens
        :param output_tokens: output tokens
        """
        self.input_tokens = input_tokens
        self.output_tokens = output_tokens
        self.total_tokens = input_tokens + output_tokens

    def get_pid(self):
        return 0

    def to_dict(self) -> dict:
        """
        Dump context item to dict

        :return: dict
        """
        return {
            "id": self.id,
            "meta_id": self.meta_id,
            "idx": self.idx,
            "first": self.first,
            "live": self.live,
            "hidden": False,
            "internal": self.internal,
            "external_id": self.external_id,
            # "stream": self.stream,  #  <-- do not dump stream response object
            "reply": self.reply,
            "current": self.current,
            "stopped": self.stopped,
            "is_audio": self.is_audio,
            "is_vision": self.is_vision,
            "agent_call": self.agent_call,
            "tool_calls": self.tool_calls,
            "index_meta": self.index_meta,
            "doc_ids": self.doc_ids,
            "sub_calls": 0,
            "sub_call": False,
            "sub_reply": False,
            "extra": self.extra,
            "extra_ctx": self.extra_ctx,
            "cmds": self.cmds,
            "cmds_before": self.cmds_before,
            "results": self.results,
            "urls": self.urls,
            "urls_before": self.urls_before,
            "images": self.images,
            "images_before": self.images_before,
            "files": self.files,
            "files_before": self.files_before,
            "attachments": self.attachments,
            "attachments_before": self.attachments_before,
            "pid": self.pid,
            "mode": self.mode,
            "model": self.model,
            "thread": self.thread,
            "msg_id": self.msg_id,
            "run_id": self.run_id,
            "audio_id": self.audio_id,
            "audio_expires_ts": self.audio_expires_ts,
            "input": self.input,
            "output": self.output,
            'hidden_input': self.hidden_input,
            'hidden_output': self.hidden_output,
            "input_name": self.input_name,
            "output_name": self.output_name,
            "input_timestamp": self.input_timestamp,
            "output_timestamp": self.output_timestamp,
            "input_tokens": self.input_tokens,
            "output_tokens": self.output_tokens,
            "total_tokens": self.total_tokens,
            "meta": self.meta.to_dict() if type(self.meta) == CtxMeta else self.meta,
        }

    def from_dict(self, data: dict):
        """
        Load context item from dict

        :param data: dict
        """
        self.id = data.get("id", None)
        self.meta = data.get("meta", None)
        self.meta_id = data.get("meta_id", None)
        self.external_id = data.get("external_id", None)
        self.stream = data.get("stream", None)
        self.cmds = data.get("cmds", [])
        self.results = data.get("results", [])
        self.urls = data.get("urls", [])
        self.urls_before = data.get("urls_before", [])
        self.images = data.get("images", [])
        self.images_before = data.get("images_before", [])
        self.files = data.get("files", [])
        self.files_before = data.get("files_before", [])
        self.attachments = data.get("attachments", [])
        self.attachments_before = data.get("attachments_before", [])
        self.reply = data.get("reply", False)
        self.input = data.get("input", None)
        self.output = data.get("output", None)
        self.mode = data.get("mode", None)
        self.model = data.get("model", None)
        self.thread = data.get("thread", None)
        self.msg_id = data.get("msg_id", None)
        self.run_id = data.get("run_id", None)
        self.audio_id = data.get("audio_id", None)
        self.audio_expires_ts = data.get("audio_expires_ts", None)
        self.input_name = data.get("input_name", None)
        self.output_name = data.get("output_name", None)
        self.input_timestamp = data.get("input_timestamp", None)
        self.output_timestamp = data.get("output_timestamp", None)
        self.hidden_input = data.get("hidden_input", None)
        self.hidden_output = data.get("hidden_output", None)
        self.input_tokens = data.get("input_tokens", 0)
        self.output_tokens = data.get("output_tokens", 0)
        self.total_tokens = data.get("total_tokens", 0)
        self.extra = data.get("extra", None)
        self.extra_ctx = data.get("extra_ctx", False)
        self.current = data.get("current", False)
        self.internal = data.get("internal", False)
        self.is_vision = data.get("is_vision", False)
        self.idx = data.get("idx", 0)
        self.first = data.get("first", False)
        self.agent_call = data.get("agent_call", False)
        self.tool_calls = data.get("tool_calls", [])
        self.index_meta = data.get("index_meta", {})
        self.doc_ids = data.get("doc_ids", [])
        self.sub_calls = data.get("sub_calls", 0)
        self.sub_call = data.get("sub_call", False)
        self.sub_reply = data.get("sub_reply", False)
        self.hidden = data.get("hidden", False)


    def dump(self) -> str:
        """
        Dump context item to JSON string

        :return: JSON string
        """
        try:
            return json.dumps(self.to_dict())
        except Exception as e:
            pass
        return ""

    def __str__(self):
        """To string"""
        return self.dump()


class CtxMeta:
    def __init__(self, id=None):
        """
        Context meta data

        :param id: Context ID
        """
        self.id = id
        self.external_id = None
        self.uuid = None
        self.name = None
        self.date = datetime.datetime.now().strftime("%Y-%m-%d")
        self.created = int(time.time())
        self.updated = int(time.time())
        self.indexed = None
        self.mode = None
        self.model = None
        self.last_mode = None
        self.last_model = None
        self.thread = None
        self.assistant = None
        self.preset = None
        self.run = None
        self.status = None
        self.extra = None
        self.initialized = False
        self.deleted = False
        self.important = False
        self.archived = False
        self.label = 0  # label color
        self.indexes = {}  # indexes data
        self.additional_ctx = [] # additional context data
        self.group = None  # parent group
        self.group_id = None
        self.root_id = None
        self.parent_id = None
        self.owner_uuid = None

    def has_additional_ctx(self) -> bool:
        """
        Check if additional context data is attached

        :return: True if additional context data is present
        """
        if self.additional_ctx is not None and len(self.additional_ctx) > 0:
            return True
        if self.group:
            if self.group.additional_ctx is not None and len(self.group.additional_ctx) > 0:
                return True
        return False

    def get_additional_ctx(self) -> list:
        """
        Get additional context data

        :return: list
        """
        if self.group:
            if self.group.additional_ctx:
                return self.group.additional_ctx
        if self.additional_ctx:
            return self.additional_ctx
        return []

    def reset_additional_ctx(self):
        """Delete additional context data"""
        if self.group:
            self.group.additional_ctx = []
        else:
            self.additional_ctx = []

    def remove_additional_ctx(self, item: dict):
        """
        Remove additional context data item

        :param item: dict
        """
        if self.group:
            if item in self.group.additional_ctx:
                self.group.additional_ctx.remove(item)
        else:
            if item in self.additional_ctx:
                self.additional_ctx.remove(item)

    def get_attachment_names(self) -> list:
        """
        Get attachment names

        :return: list
        """
        if self.group:
            if self.group.additional_ctx:
                return self.group.get_attachment_names()
        return [item['name'] for item in self.additional_ctx]

    def to_dict(self) -> dict:
        """
        Dump context meta to dict

        :return: dict
        """
        data = {
            "id": self.id,
            "external_id": self.external_id,
            "uuid": self.uuid,
            "name": self.name,
            "date": self.date,
            "created": self.created,
            "updated": self.updated,
            "indexed": self.indexed,
            "mode": self.mode,
            "model": self.model,
            "last_mode": self.last_mode,
            "last_model": self.last_model,
            "thread": self.thread,
            "assistant": self.assistant,
            "preset": self.preset,
            "run": self.run,
            "status": self.status,
            "extra": self.extra,
            "initialized": self.initialized,
            "deleted": self.deleted,
            "important": self.important,
            "archived": self.archived,
            "label": self.label,
            "indexes": self.indexes,
            "additional_ctx": self.additional_ctx,
            "group_id": self.group_id,
            "root_id": self.root_id,
            "parent_id": self.parent_id,
            "owner_uuid": self.owner_uuid,
        }
        if self.group:
            data["__group__"] = self.group.to_dict()
        return data

    def from_dict(self, data: dict):
        """
        Load context meta from dict

        :param data: dict
        """
        self.id = data.get("id", None)
        self.external_id = data.get("external_id", None)
        self.uuid = data.get("uuid", None)
        self.name = data.get("name", None)
        self.date = data.get("date", None)
        self.created = data.get("created", None)
        self.updated = data.get("updated", None)
        self.indexed = data.get("indexed", None)
        self.mode = data.get("mode", None)
        self.model = data.get("model", None)
        self.last_mode = data.get("last_mode", None)
        self.last_model = data.get("last_model", None)
        self.thread = data.get("thread", None)
        self.assistant = data.get("assistant", None)
        self.preset = data.get("preset", None)
        self.run = data.get("run", None)
        self.status = data.get("status", None)
        self.extra = data.get("extra", None)
        self.initialized = data.get("initialized", False)
        self.deleted = data.get("deleted", False)
        self.important = data.get("important", False)
        self.archived = data.get("archived", False)
        self.label = data.get("label", 0)
        self.indexes = data.get("indexes", {})
        self.additional_ctx = data.get("additional_ctx", [])
        self.group_id = data.get("group_id", None)
        self.root_id = data.get("root_id", None)
        self.parent_id = data.get("parent_id", None)
        self.owner_uuid = data.get("owner_uuid", None)

    def get_pid(self):
        return 0

class CtxGroup:
    def __init__(self, id=None, name=None):
        """
        Context group

        :param id: Group ID
        :param name: Group name
        """
        self.id = id
        self.uuid = None
        self.name = name
        self.items = []
        self.created = int(time.time())
        self.updated = int(time.time())
        self.additional_ctx = []
        self.count = 0

    def has_additional_ctx(self) -> bool:
        """
        Check if additional context data is attached

        :return: True if additional context data is present
        """
        if self.additional_ctx is not None and len(self.additional_ctx) > 0:
            return True
        return False

    def get_additional_ctx(self) -> list:
        """
        Get additional context data

        :return: list
        """
        return self.additional_ctx

    def get_attachments_count(self) -> int:
        """
        Get attachments count

        :return: int
        """
        return len(self.additional_ctx)

    def get_attachment_names(self) -> list:
        """
        Get attachment names

        :return: list
        """
        return [item['name'] for item in self.additional_ctx]

    def to_dict(self) -> dict:
        """
        Dump context group to dict

        :return: dict
        """
        return {
            "id": self.id,
            "uuid": self.uuid,
            "name": self.name,
            "items": self.items,
            "additional_ctx": self.additional_ctx,
            "created": self.created,
            "updated": self.updated,
            "count": self.count,
        }

    def from_dict(self, data: dict):
        """
        Load context group from dict

        :param data: dict
        """
        self.id = data.get("id", None)
        self.uuid = data.get("uuid", None)
        self.name = data.get("name", None)
        self.items = data.get("items", [])
        self.created = data.get("created", None)
        self.updated = data.get("updated", None)
        self.count = data.get("count", 0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.plugin import BasePlugin
from pygpt_net.core.events import Event
from pygpt_net.item.ctx import CtxItem
from .config import Config


class Plugin(BasePlugin):
    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.id = "agent"
        self.name = "Autonomous Agent (inline)"
        self.description = "Enables inline autonomous mode (Agent) in current mode. " \
                           "WARNING: Please use with caution - this mode, when connected with other plugins, " \
                           "may produce unexpected results!"
        self.prefix = "Agent"
        self.type = [
            "agent",
            "cmd.inline",
        ]
        self.order = 9998
        self.use_locale = True
        self.config = Config(self)
        self.init_options()

    def init_options(self):
        """Initialize options"""
        self.config.from_defaults(self)

    def is_allowed(self) -> bool:
        """
        Check if plugin is allowed to run

        :return: True if plugin is allowed to run
        """
        return self.window.core.config.get("mode") != "agent"  # check global mode

    def handle(self, event: Event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        :param args: event args
        :param kwargs: event kwargs
        """
        name = event.name
        data = event.data
        ctx = event.ctx

        always_events = [
            Event.FORCE_STOP,
            Event.PLUGIN_SETTINGS_CHANGED,
            Event.ENABLE,
            Event.DISABLE
        ]

        if not self.is_allowed() and name != Event.DISABLE:
            return

        if name not in always_events and not self.is_active_prompt():
            return

        if name == Event.CTX_BEFORE:
            self.on_ctx_before(ctx)

        elif name == Event.CTX_AFTER:
            self.on_ctx_after(ctx)

        elif name == Event.CTX_END:
            self.on_ctx_end(ctx)

        elif name == Event.USER_SEND:
            self.on_user_send(data['value'])

        elif name == Event.FORCE_STOP:
            self.on_stop()

        elif name == Event.SYSTEM_PROMPT:
            data['value'] = self.on_system_prompt(data['value'])

        elif name == Event.INPUT_BEFORE:
            data['value'] = self.on_input_before(data['value'])

        elif name in [
            Event.ENABLE,
            Event.DISABLE,
        ]:
            if data['value'] == self.id:
                self.window.controller.agent.legacy.update()  # update agent status bar

        elif name == Event.PLUGIN_SETTINGS_CHANGED:
            self.window.controller.agent.legacy.update()  # update agent status bar

        elif name in [
            Event.CMD_INLINE,
            Event.CMD_EXECUTE,
        ]:
            if self.get_option_value("auto_stop"):
                self.cmd(
                    ctx,
                    data['commands'],
                )

    def is_active_prompt(self) -> bool:
        """
        Check if active prompt is enabled

        :return: True if active prompt is enabled
        """
        for item in self.get_option_value("prompts"):
            if item["enabled"]:
                return True
        return False

    def on_system_prompt(self, prompt: str) -> str:
        """
        Event: SYSTEM_PROMPT

        :param prompt: prompt
        :return: updated prompt
        """
        pre_prompt = ("YOU ARE NOW AN AUTONOMOUS AGENT AND YOU ARE ENTERING NOW INTO AGENT MODE.\n"
                      "Use below instructions in every agent run iteration:\n\n")
        return pre_prompt + self.window.controller.agent.legacy.on_system_prompt(
            prompt,
            append_prompt=self.get_first_active_prompt(),
            auto_stop=self.get_option_value("auto_stop"),
        )

    def on_input_before(self, prompt: str) -> str:
        """
        Event: INPUT_BEFORE

        :param prompt: prompt
        :return: updated prompt
        """
        return self.window.controller.agent.legacy.on_input_before(prompt)

    def cmd(self, ctx: CtxItem, cmds: list):
        """
        Events: CMD_INLINE, CMD_EXECUTE

        :param ctx: CtxItem
        :param cmds: commands dict
        """
        self.window.controller.agent.legacy.cmd(ctx, cmds)  # force execute

    def on_stop(self):
        """
        Event: FORCE_STOP
        """
        self.window.controller.agent.legacy.on_stop()  # force stop

    def on_user_send(self, text: str):
        """
        Event: USER_SEND

        :param text: text
        """
        self.window.controller.agent.legacy.on_user_send(text)

    def on_ctx_end(self, ctx: CtxItem):
        """
        Event: CTX_END

        :param ctx: CtxItem
        """
        self.window.controller.agent.legacy.on_ctx_end(
            ctx,
            iterations=int(self.get_option_value("iterations")),
        )

    def on_ctx_before(self, ctx: CtxItem):
        """
        Event: CTX_BEFORE

        :param ctx: CtxItem
        """
        self.window.controller.agent.legacy.on_ctx_before(
            ctx,
            reverse_roles=self.get_option_value("reverse_roles"),
        )

    def on_ctx_after(self, ctx: CtxItem):
        """
        Event: CTX_AFTER

        :param ctx: CtxItem
        """
        self.window.controller.agent.legacy.on_ctx_after(ctx)

    def get_first_active_prompt(self) -> str:
        """
        Get first active prompt from prompts list

        :return: system prompt
        """
        for item in self.get_option_value("prompts"):
            if item["enabled"]:
                return item["prompt"]
        return ""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import json
from typing import Optional, Dict, Any

from pygpt_net.item.ctx import CtxItem

class BaseEvent:
    def __init__(
            self,
            name: Optional[str] = None,
            data: Optional[dict] = None,
            ctx: CtxItem = None
    ):
        """
        Base Event object class

        :param name: event name
        :param data: event data
        :param ctx: context instance
        """
        self.id = None
        self.name = name
        self.data = data
        if self.data is None:
            self.data = {}
        self.ctx = ctx  # CtxItem
        self.stop = False  # True to stop propagation
        self.internal = False
        self.call_id = 0


    @property
    def full_name(self) -> str:
        """
        Get full event name

        :return: Full event name
        """
        return self.id + ": " + self.name

    def to_dict(self) -> Dict[str, Any]:
        """Dump event to dict"""
        return {
            'name': self.name,
            'data': self.data,
            'stop': self.stop,
            'internal': self.internal,
        }

    def dump(self) -> str:
        """
        Dump event to json string

        :return: JSON string
        """
        try:
            return json.dumps(self.to_dict())
        except Exception as e:
            pass
        return ""

    def __str__(self):
        return self.dump()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.19 02:00:00                  #
# ================================================== #

import os
from typing import Dict

from PySide6.QtCore import QTimer
from PySide6.QtGui import QTextCursor, QAction, QIcon
from PySide6.QtWidgets import QWidget

from pygpt_net.core.tabs.tab import Tab
from pygpt_net.tools.base import BaseTool
from pygpt_net.tools.code_interpreter.ui.dialogs import Tool
from pygpt_net.core.events import Event
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans

from .ui.widgets import PythonInput, PythonOutput, ToolWidget, ToolSignals


class CodeInterpreter(BaseTool):
    def __init__(self, *args, **kwargs):
        """
        Python code interpreter

        :param window: Window instance
        """
        super(CodeInterpreter, self).__init__(*args, **kwargs)
        self.id = "interpreter"
        self.has_tab = True
        self.tab_title = "menu.tools.interpreter"
        self.tab_icon = ":/icons/code.svg"
        self.opened = False
        self.is_edit = False
        self.auto_clear = False
        self.dialog = None
        self.ipython = True
        self.signals = ToolSignals()

        # interpreter data files in /data directory
        self.file_current = ".interpreter.current.py"
        self.file_input = ".interpreter.input.py"
        self.file_output = ".interpreter.output.py"

    def setup(self):
        """Setup"""
        self.load_history()
        self.load_output()
        self.update()

        # restore
        if self.window.core.config.has("interpreter.input"):
            self.signals.update_input.emit(self.window.core.config.get("interpreter.input"))
        if self.window.core.config.has("interpreter.execute_all"):
            self.signals.set_checkbox_all.emit(self.window.core.config.get("interpreter.execute_all"))
        if self.window.core.config.has("interpreter.auto_clear"):
            self.signals.set_checkbox_auto_clear.emit(self.window.core.config.get("interpreter.auto_clear"))
        if self.window.core.config.has("interpreter.ipython"):
            self.signals.set_checkbox_ipython.emit(self.window.core.config.get("interpreter.ipython"))
        if self.ipython:
            self.signals.toggle_all_visible.emit(False)

        # set initial size
        if not self.window.core.config.has("interpreter.dialog.initialized"):
            self.set_initial_size()
            self.window.core.config.set("interpreter.dialog.initialized", True)

    def set_initial_size(self):
        """Set default sizes"""
        def set_initial_splitter_height():
            total_height = self.window.ui.splitters['interpreter'].size().height()
            if total_height > 0:
                size_output = int(total_height * 0.85)
                size_input = total_height - size_output
                self.window.ui.splitters['interpreter'].setSizes([size_output, size_input])
            else:
                QTimer.singleShot(0, set_initial_splitter_height)
        QTimer.singleShot(0, set_initial_splitter_height)

        def set_initial_splitter_width():
            total_width = self.window.ui.splitters['interpreter.columns'].size().width()
            if total_width > 0:
                size_output = int(total_width * 0.85)
                size_history = total_width - size_output
                self.window.ui.splitters['interpreter.columns'].setSizes([size_output, size_history])
            else:
                QTimer.singleShot(0, set_initial_splitter_width)
        QTimer.singleShot(0, set_initial_splitter_width)

    def handle_ipython_output(self, line: str):
        """
        Handle ipython output

        :param line: Output line
        """
        self.append_output(line)

    def on_reload(self):
        """On app profile reload"""
        self.setup()

    def update(self):
        """Update menu"""
        self.update_menu()

    def update_menu(self):
        """Update menu"""
        """
        if self.opened:
            self.window.ui.menu['tools.interpreter'].setChecked(True)
        else:
            self.window.ui.menu['tools.interpreter'].setChecked(False)
        """

    def append_output(self, output: str, type="stdout", **kwargs):
        """
        Append output to the interpreter window

        :param output: Output data
        :param type: Output type
        :param kwargs: Additional parameters
        """
        self.signals.update.emit(output, type)
        self.save_output()
        self.load_history()

    def get_path_input(self) -> str:
        """
        Get input path

        :return: Input path
        """
        return os.path.join(self.window.core.config.get_user_dir("data"), self.file_input)

    def get_path_output(self) -> str:
        """
        Get output path

        :return: Output path
        """
        return os.path.join(self.window.core.config.get_user_dir("data"), self.file_output)

    def get_widget(self) -> ToolWidget:
        """
        Get tool widget

        :return: ToolWidget instance
        """
        return self.dialog.widget

    def get_widget_history(self) -> PythonOutput:
        """
        Get history widget

        :return: PythonOutput widget
        """
        return self.dialog.widget.history

    def get_widget_output(self) -> PythonOutput:
        """
        Get output widget

        :return: PythonOutput widget
        """
        return self.dialog.widget.output

    def get_widget_input(self) -> PythonInput:
        """
        Get input widget

        :return: PythonInput widget
        """
        return self.dialog.widget.input

    def load_history(self):
        """Load history data from file"""
        data = self.get_history()
        self.signals.update_history.emit(data)

    def get_output(self) -> str:
        """
        Load and get output from file

        :return: Output data
        """
        data = ""
        path = self.get_path_output()
        if os.path.exists(path):
            with open(path, "r", encoding="utf-8") as f:
                try:
                    data = f.read()
                except:
                    pass
        return data

    def load_output(self):
        """Load output data from file"""
        data = self.get_output()
        self.signals.update.emit(data, "stdout")
        self.signals.focus_input.emit()

    def save_output(self):
        """Save output data to file"""
        path = self.get_path_output()
        data = self.get_widget_output().toPlainText()
        with open(path, "w", encoding="utf-8") as f:
            f.write(data)

    def get_history(self) -> str:
        """
        Load and get history from file

        :return: History data
        """
        data = ""
        path = self.get_path_input()
        if os.path.exists(path):
            with open(path, "r", encoding="utf-8") as f:
                try:
                    data = f.read()
                except:
                    pass
        return data

    def save_history(self, input: str):
        """
        Save input data

        :param input: Input data
        """
        path = self.get_path_input()
        with open(path , "w", encoding="utf-8") as f:
            f.write(input)

    def clear_history(self):
        """Clear input"""
        path = self.get_path_input()
        if os.path.exists(path):
            os.remove(path)
        self.signals.clear_history.emit()

    def clear_output(self):
        """Clear output"""
        path = self.get_path_output()
        if os.path.exists(path):
            os.remove(path)
        self.signals.clear_output.emit()

    def clear(self, force: bool = False):
        """
        Clear current window

        :param force: Force clear
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type='interpreter.clear',
                id=0,
                msg=trans("interpreter.clear.confirm"),
            )
            return
        self.clear_output()

    def clear_all(self):
        """Clear input and output"""
        self.clear_output()
        self.clear_history()

    def restart_kernel(self):
        """Restart kernel"""
        print("Sending restart event...")
        commands = [
            {
                "cmd": "ipython_kernel_restart",
                "params": {},
                "silent": True,
                "force": True,
            }
        ]
        event = Event(Event.CMD_EXECUTE, {
            'commands': commands,
            'silent': True,
        })
        event.ctx = CtxItem()  # tmp
        self.window.controller.command.dispatch_only(event)
        self.signals.focus_input.emit()

    def send_input(self, widget: ToolWidget):
        """
        Send input to interpreter

        :param widget: PythonInput widget
        """
        self.store_history(widget)
        if self.auto_clear:
            self.clear_output()

        input_textarea = widget.input
        input = str(input_textarea.toPlainText()).strip()

        if input == "/restart":
            self.restart_kernel()
            input_textarea.clear()
            input_textarea.setFocus()
            return
        elif input == "/clear":
            self.clear(force=True)
            input_textarea.clear()
            input_textarea.setFocus()
            return

        if self.is_all():
            cmd = "code_execute_all"
        else:
            if input == "":
                return
            cmd = "code_execute"

        if self.ipython:
            cmd = "ipython_execute"

        commands = [
            {
                "cmd": cmd,
                "params": {
                    "code": input,
                    "path": self.file_current,
                },
                "silent": True,
                "force": True,
            }
        ]
        event = Event(Event.CMD_EXECUTE, {
            'commands': commands,
            'silent': True,
        })
        event.ctx = CtxItem()  # tmp
        self.window.controller.command.dispatch_only(event)
        input_textarea.clear()
        input_textarea.setFocus()

    def update_input(self):
        """Update input data"""
        data = self.get_widget_input().toPlainText()
        self.window.core.config.set("interpreter.input", data)

    def append_to_input(self, data: str):
        """
        Append data to input

        :param data: Data
        """
        self.signals.append_input.emit(data)

    def append_to_edit(self, data: str):
        """
        Append data to editor

        :param data: Data
        """
        prev = self.get_history()
        if prev:
            prev += "\n"
        prev += data
        self.save_history(prev)
        self.load_history()

    def store_history(self, widget: ToolWidget):
        """
        Save history data to file

        :param widget: ToolWidget
        """
        if not widget.history:
            return
        data = widget.history.toPlainText()
        self.save_history(data)

    def open(self):
        """Open interpreter dialog"""
        self.opened = True
        self.load_history()
        self.load_output()
        self.window.ui.dialogs.open('interpreter', width=800, height=600)
        self.get_widget_input().setFocus()
        self.cursor_to_end()
        self.scroll_to_bottom()
        self.update()

    def close(self):
        """Close interpreter dialog"""
        self.opened = False
        self.window.ui.dialogs.close('interpreter')
        self.update()

    def scroll_to_bottom(self):
        """Scroll down"""
        self.get_widget_history().verticalScrollBar().setValue(
            self.get_widget_history().verticalScrollBar().maximum()
        )
        self.get_widget_output().verticalScrollBar().setValue(
            self.get_widget_output().verticalScrollBar().maximum()
        )

    def toggle(self):
        """Toggle interpreter dialog open/close"""
        if self.opened:
            self.close()
        else:
            self.open()

    def show_hide(self, show: bool = True):
        """
        Show/hide interpreter window

        :param show: show/hide
        """
        if show:
            self.open()
        else:
            self.close()

    def get_toolbar_icon(self) -> QWidget:
        """
        Get toolbar icon

        :return: QWidget
        """
        return self.window.ui.nodes['icon.interpreter']

    def toggle_icon(self, state: bool):
        """
        Toggle interpreter icon

        :param state: State
        """
        self.get_toolbar_icon().setVisible(state)

    def toggle_auto_clear(self, widget: ToolWidget):
        """
        Toggle auto clear

        :param widget: ToolWidget instance
        """
        self.auto_clear = widget.checkbox_auto_clear.isChecked()
        self.window.core.config.set("interpreter.auto_clear", self.auto_clear)
        self.signals.set_checkbox_auto_clear.emit(self.auto_clear)

    def toggle_ipython(self, widget: ToolWidget):
        """
        Toggle ipython

        :param widget: ToolWidget instance
        """
        self.ipython = widget.checkbox_ipython.isChecked()
        self.window.core.config.set("interpreter.ipython", self.ipython)
        self.signals.set_checkbox_ipython.emit(self.ipython)
        if self.ipython:
            self.signals.toggle_all_visible.emit(False)
        else:
            self.signals.toggle_all_visible.emit(True)

    def toggle_all(self, widget: ToolWidget):
        """
        Toggle execute all

        :param widget: ToolWidget instance
        """
        state = widget.checkbox_all.isChecked()
        self.window.core.config.set("interpreter.execute_all", state)
        self.signals.set_checkbox_all.emit(state)

    def get_current_output(self) -> str:
        """
        Get current output from interpreter

        :return: Output data
        """
        return self.get_output()

    def get_current_history(self) -> str:
        """
        Get code from history

        :return: Edit code
        """
        return self.get_history()

    def is_all(self) -> bool:
        """
        Check if execute all is enabled

        :return: True if execute all is enabled
        """
        return self.get_widget().checkbox_all.isChecked()

    def cursor_to_end(self):
        """Move cursor to end"""
        cur = self.get_widget_history().textCursor()
        cur.movePosition(QTextCursor.End)
        self.get_widget_history().setTextCursor(cur)

    def setup_menu(self) -> Dict[str, QAction]:
        """
        Setup main menu

        :return dict with menu actions
        """
        actions = {}
        actions["interpreter"] = QAction(
            QIcon(":/icons/code.svg"),
            trans("menu.tools.interpreter"),
            self.window,
            checkable=False,
        )
        actions["interpreter"].triggered.connect(
            lambda: self.toggle()
        )
        return actions

    def as_tab(self, tab: Tab) -> QWidget:
        """
        Spawn and return tab instance

        :param tab: Parent Tab instance
        :return: Tab widget instance
        """
        tool = Tool(window=self.window, tool=self)
        layout = tool.widget.setup(all=False)
        widget = QWidget()
        widget.setLayout(layout)
        self.load_history()
        self.load_output()
        tool.set_tab(tab)
        return widget

    def setup_dialogs(self):
        """Setup dialogs (static)"""
        self.dialog = Tool(self.window, self)
        self.dialog.setup()

    def setup_theme(self):
        """Setup theme"""
        self.get_widget_output().value = self.window.core.config.get('font_size')

    def get_lang_mappings(self) -> Dict[str, Dict]:
        """
        Get language mappings

        :return: dict with language mappings
        """
        return {
            'menu.text': {
                'tools.interpreter': 'menu.tools.interpreter',
            }
        }


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 00:00:00                  #
# ================================================== #

from typing import Optional

from .plugin import BasePlugin

class BaseConfig:
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        self.plugin = plugin

    def from_defaults(self, plugin: Optional[BasePlugin] = None):
        """
        Set default options for plugin

        :param plugin: plugin instance
        """
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.24 06:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        """
        Set default options for plugin

        :param plugin: plugin instance
        """
        dockerfile = 'FROM python:3.9-alpine'
        dockerfile += '\n\n'
        dockerfile += 'RUN mkdir /data'
        dockerfile += '\n\n'
        dockerfile += '# Data directory, bound as a volume to the local \'data/\' directory'
        dockerfile += '\nWORKDIR /data'

        volumes_keys = {
            "enabled": "bool",
            "docker": "text",
            "host": "text",
        }
        volumes_items = [
            {
                "enabled": True,
                "docker": "/data",
                "host": "{workdir}",
            },
        ]
        ports_keys = {
            "enabled": "bool",
            "docker": "text",
            "host": "int",
        }
        ports_items = []

        plugin.add_option(
            "sandbox_docker",
            type="bool",
            value=False,
            label="Sandbox (docker container)",
            description="Executes commands in sandbox (docker container). "
                        "Docker must be installed and running.",
            tab="sandbox",
        )
        plugin.add_option(
            "dockerfile",
            type="textarea",
            value=dockerfile,
            label="Dockerfile",
            description="Dockerfile",
            tooltip="Dockerfile",
            tab="sandbox",
        )
        plugin.add_option(
            "image_name",
            type="text",
            value='pygpt_system',
            label="Docker image name",
            tab="sandbox",
        )
        plugin.add_option(
            "container_name",
            type="text",
            value='pygpt_system_container',
            label="Docker container name",
            tab="sandbox",
        )
        plugin.add_option(
            "docker_entrypoint",
            type="text",
            value='tail -f /dev/null',
            label="Docker run command",
            tab="sandbox",
            advanced=True,
        )
        plugin.add_option(
            "docker_volumes",
            type="dict",
            value=volumes_items,
            label="Docker volumes",
            description="Docker volumes mapping",
            tooltip="Docker volumes mapping",
            keys=volumes_keys,
            tab="sandbox",
            advanced=True,
        )
        plugin.add_option(
            "docker_ports",
            type="dict",
            value=ports_items,
            label="Docker ports",
            description="Docker ports mapping",
            tooltip="Docker ports mapping",
            keys=ports_keys,
            tab="sandbox",
            advanced=True,
        )
        plugin.add_option(
            "auto_cwd",
            type="bool",
            value=True,
            label="Auto-append CWD to sys_exec",
            description="Automatically append current working directory to sys_exec command",
            tab="general",
        )
        plugin.add_cmd(
            "sys_exec",
            instruction="execute ANY system command, script or app in user's environment. "
                        "Do not use this command to install Python libraries, use IPython environment and IPython commands instead.",
            params=[
                {
                    "name": "command",
                    "type": "str",
                    "description": "system command",
                    "required": True,
                },
            ],
            enabled=True,
            description="Allows system commands execution",
            tab="general",
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.25 02:00:00                  #
# ================================================== #

import re
from jupyter_client import KernelManager

class LocalKernel:
    def __init__(self, plugin = None):
        self.plugin = plugin
        self.client = None
        self.manager = None
        self.initialized = False
        self.signals = None

    def restart_kernel(self) -> bool:
        """
        Restart the IPython kernel.

        :return: True if successful.
        """
        if self.initialized:
            self.manager.stop_channels()
            self.manager.restart_kernel(now=True)
            self.manager = self.manager.client()
            self.manager.start_channels()
            self.log("IPython kernel restarted.")
        else:
            self.init()
            self.log("IPython kernel started.")
        return True

    def shutdown_kernel(self):
        """Shutdown the IPython kernel."""
        self.manager.stop_channels()
        self.manager.shutdown_kernel()

    def init(self, force: bool = False):
        """
        Initialize the IPython kernel client.

        :param force: Force reinitialization.
        """
        if self.initialized and not force:
            return
        self.manager = KernelManager()
        self.manager.start_kernel()
        self.client = self.manager.client()
        self.client.start_channels()
        self.client.wait_for_ready()
        self.log("Connected to local IPython kernel.")
        self.initialized = True
        self.log("IPython kernel is ready.")

    def process_message(self, msg: dict) -> str:
        """
        Process the message from the IPython kernel.

        :param msg: Message from the IPython kernel.
        :return: Processed message.
        """
        msg_type = msg['msg_type']
        content = msg['content']
        if msg_type == 'stream':
            # standard output and error
            return content['text']
        elif msg_type == 'display_data':
            # display data
            data = content['data']
            if 'text/plain' in data:
                return data['text/plain']
        elif msg_type == 'execute_result':
            # execution result
            data = content['data']
            if 'text/plain' in data:
                return data['text/plain']
        elif msg_type == 'error':
            # execution errors
            return "Error executing code:" + "\n".join(content['traceback'])
        return ""

    def end(self, all: bool = False):
        """
        Stop the IPython kernel.

        :param all: Stop the container as well.
        """
        self.client.stop_channels()  # stop the client
        self.manager.shutdown_kernel()

    def execute(self, code: str, current: bool = False) -> str:
        """
        Execute the code in the IPython kernel.

        :param code: Python code to execute.
        :param current: Use the current kernel.
        :return: Output from the kernel.
        """
        self.init()
        if not current:
            self.restart_kernel()

        self.log("Executing code: " + str(code)[:100] + "...")

        msg_id = self.client.execute(code)
        output = ""
        while True:
            try:
                msg = self.client.get_iopub_msg(timeout=1)
            except:
                continue

            if msg['parent_header'].get('msg_id') != msg_id:
                continue

            chunk = str(self.process_message(msg))
            if chunk.strip() != "":
                output += chunk
                self.send_output(chunk)

            if (msg['msg_type'] == 'status' and
                    msg['content']['execution_state'] == 'idle'):
                break

        return self.remove_ansi(output).strip()

    def send_output(self, output: str):
        """
        Send the output to the output.

        :param output: Output.
        :return: Output.
        """
        if self.signals is not None:
            self.signals.ipython_output.emit(output)

    def remove_ansi_more(self, text):
        """
        Clean the text from ANSI escape sequences, carriage returns, and progress bars.

        :param text: Text to clean.
        :return: Cleaned text.
        """
        # Remove ANSI escape sequences
        ansi_escape = re.compile(r'''
            \x1B  # ESC
            (?:   # Start of sequence
                [@-Z\\-_]  # ESC [@ to ESC _ (7-bit C1 Control codes)
            |     # or
                \[  # ESC[
                [0-?]*  # Parameter bytes
                [ -/]*  # Intermediate bytes
                [@-~]   # Final byte
            )
        ''', re.VERBOSE)
        text = ansi_escape.sub('', text)

        # Split text into lines
        lines = text.split('\n')

        cleaned_lines = []
        for line in lines:
            # Handle carriage returns - keep only the text after the last '\r'
            if '\r' in line:
                line = line.split('\r')[-1]
            # Optionally, remove progress bar lines by detecting lines containing box-drawing characters
            # Skip the line if it contains box-drawing characters
            if re.search(r'[\u2500-\u257F]', line):
                continue  # Skip this line
            # Append the cleaned line
            cleaned_lines.append(line)
        # Reconstruct the text
        text = '\n'.join(cleaned_lines)
        return text

    def remove_ansi(self, text) -> str:
        """
        Clean the text from ANSI escape sequences.

        :param text: Text to clean.
        :return: Cleaned text.
        """
        ansi_escape = re.compile(
            r'''
            \x1B   # ESC
            (?:    # 7-bit C1 Fe
                [@-Z\\-_]
            |      # or 8-bit C1 Fe
                \[
                [0-?]*   # Parameter bytes
                [ -/]*   # Intermediate bytes
                [@-~]    # Final byte
            )
            ''',
            re.VERBOSE
        )
        return ansi_escape.sub('', text)

    def attach_signals(self, signals):
        """
        Attach signals

        :param signals: signals
        """
        self.signals = signals

    def log(self, msg):
        """
        Log the message.

        :param msg: Message to log.
        """
        print(msg)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.29 23:00:00                  #
# ================================================== #

from PySide6.QtCore import QObject, Signal, QRunnable, Slot

from pygpt_net.core.types import (
    MODE_AGENT_LLAMA,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
)
from pygpt_net.core.events import KernelEvent, Event


class BridgeSignals(QObject):
    """Bridge signals"""
    response = Signal(object)  # KernelEvent


class BridgeWorker(QObject, QRunnable):
    """Bridge worker"""
    def __init__(self, *args, **kwargs):
        QObject.__init__(self)
        QRunnable.__init__(self)
        self.signals = BridgeSignals()
        self.args = args
        self.kwargs = kwargs
        self.window = None
        self.context = None
        self.extra = None
        self.mode = None

    @Slot()
    def run(self):
        """Run bridge worker"""
        self.window.core.debug.info("[bridge] Worker started.")
        result = False

        try:
            # POST PROMPT ASYNC: handle post prompt async event
            self.handle_post_prompt_async()

            # ADDITIONAL CONTEXT: append additional context from attachments
            self.handle_additional_context()

            # POST PROMPT END: handle post prompt end event
            self.handle_post_prompt_end()

            # Langchain
            if self.mode == MODE_LANGCHAIN:
                result = self.window.core.chain.call(
                    context=self.context,
                    extra=self.extra,
                )

            # LlamaIndex: chat with files
            elif self.mode == MODE_LLAMA_INDEX:
                result = self.window.core.idx.chat.call(
                    context=self.context,
                    extra=self.extra,
                )

            # LlamaIndex: agents
            elif self.mode == MODE_AGENT_LLAMA:
                result = self.window.core.agents.runner.call(
                    context=self.context,
                    extra=self.extra,
                    signals=self.signals,
                )
                if result:
                    return  # don't emit any signals (handled in agent runner, step by step)
                else:
                    self.extra["error"] = str(self.window.core.agents.runner.get_error())

            # Loop: next step
            elif self.mode == "loop_next":  # virtual mode
                result = self.window.core.agents.runner.run_next(
                    context=self.context,
                    extra=self.extra,
                    signals=self.signals,
                )
                if result:
                    return  # don't emit any signals (handled in agent runner, step by step)
                else:
                    self.extra["error"] = str(self.window.core.agents.runner.get_error())

            # API OpenAI: chat, completion, vision, image, assistants
            else:
                result = self.window.core.gpt.call(
                    context=self.context,
                    extra=self.extra,
                )
        except Exception as e:
            if self.signals:
                self.extra["error"] = e
                event = KernelEvent(KernelEvent.RESPONSE_FAILED, {
                    'context': self.context,
                    'extra': self.extra,
                })
                self.signals.response.emit(event)
                return

        # send response to main thread
        if self.signals:
            name = KernelEvent.RESPONSE_OK if result else KernelEvent.RESPONSE_ERROR
            event = KernelEvent(name, {
                'context': self.context,
                'extra': self.extra,
            })
            self.signals.response.emit(event)

    def handle_post_prompt_async(self):
        """Handle post prompt async event"""
        event = Event(Event.POST_PROMPT_ASYNC, {
            'mode': self.context.mode,
            'reply': self.context.ctx.reply,
            'value': self.context.system_prompt,
        })
        event.ctx = self.context.ctx
        self.window.dispatch(event)
        self.context.system_prompt = event.data['value']

    def handle_post_prompt_end(self):
        """Handle post prompt end event"""
        event = Event(Event.POST_PROMPT_END, {
            'mode': self.context.mode,
            'reply': self.context.ctx.reply,
            'value': self.context.system_prompt,
        })
        event.ctx = self.context.ctx
        self.window.dispatch(event)
        self.context.system_prompt = event.data['value']

    def handle_additional_context(self):
        """Append additional context"""
        ctx = self.context.ctx
        if ctx is None:
            return
        if ctx.meta is None:
            return
        if not self.window.controller.chat.attachment.has_context(ctx.meta):
            return
        ad_context = self.window.controller.chat.attachment.get_context(ctx, self.context.history)
        ad_mode = self.window.controller.chat.attachment.get_mode()
        if ad_context:
            self.context.prompt += "\n\n" + ad_context  # append to input text
            if (ad_mode == self.window.controller.chat.attachment.MODE_QUERY_CONTEXT
                    or self.mode == MODE_AGENT_LLAMA):
                ctx.hidden_input = ad_context  # store for future use, only if query context
                # if full context or summary, then whole extra context will be applied to current input



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.18 23:00:00                  #
# ================================================== #

from typing import Optional, List, Dict, Any

import os

from PySide6.QtCore import QTimer, Slot
from PySide6.QtWidgets import QApplication

from pygpt_net.item.ctx import CtxItem
from pygpt_net.plugin.audio_input.worker import ControlWorker
from pygpt_net.core.events import ControlEvent, AppEvent
from pygpt_net.utils import trans


class Voice:

    MIN_FRAMES = 25  # minimum frames to start transcription
    PLAY_DELAY = 500  # ms, delay before playing audio event

    def __init__(self, window=None):
        """
        Accessibility voice controller

        :param window: Window instance
        """
        self.window = window
        self.is_recording = False
        self.timer = None
        self.input_file = "voice_control.wav"
        self.thread_started = False
        self.audio_disabled_events = [
            AppEvent.INPUT_CALL,
            AppEvent.VOICE_CONTROL_TOGGLE,
        ]
        self.confirm_events = [
            ControlEvent.NOTEPAD_CLEAR,
            ControlEvent.CALENDAR_CLEAR,
        ]
        self.pending_text = None
        self.pending_event = None
        self.play_timer = QTimer(self.window)
        self.play_timer.timeout.connect(self.play_audio)

    def setup(self):
        """Setup voice control"""
        self.update()

    def delayed_play(self, text: str, event: AppEvent):
        """
        Delayed play audio

        :param text: text to play
        :param event: AppEvent
        """
        self.pending_text = text
        self.pending_event = event
        self.play_timer.start(self.PLAY_DELAY)

    def play_audio(self):
        """Play current pending audio"""
        self.play_timer.stop()
        if self.pending_text is not None and self.pending_event is not None:
            self.window.controller.audio.play_event(
                self.pending_text,
                self.pending_event,
            )  # use cached audio if available
            self.pending_text = None
            self.pending_event = None

    def play(self, event: AppEvent):
        """
        Play audio event

        :param event: AppEvent
        """
        mic_events = [
            AppEvent.VOICE_CONTROL_STARTED,
            AppEvent.VOICE_CONTROL_STOPPED,
            AppEvent.VOICE_CONTROL_SENT,
            AppEvent.INPUT_VOICE_LISTEN_STARTED,
            AppEvent.INPUT_VOICE_LISTEN_STOPPED,
        ]
        always_play = [
            AppEvent.VOICE_CONTROL_UNRECOGNIZED,
        ]
        if event.name in mic_events and self.window.core.config.get("access.microphone.notify"):
            if event.name in [AppEvent.VOICE_CONTROL_STARTED, AppEvent.INPUT_VOICE_LISTEN_STARTED]:
                if self.window.core.config.get("access.audio.notify.execute"):
                    self.window.controller.audio.play_sound("click_on.mp3")
            elif event.name in [
                AppEvent.VOICE_CONTROL_STOPPED,
                AppEvent.VOICE_CONTROL_SENT,
                AppEvent.INPUT_VOICE_LISTEN_STOPPED,
            ]:
                if self.window.core.config.get("access.audio.notify.execute"):
                    self.window.controller.audio.play_sound("click_off.mp3")
            return

        if not self.window.core.config.get("access.audio.event.speech") and event.name not in always_play:
            return

        if event.name not in self.audio_disabled_events:
            trans_key = "event.audio." + event.name

            if event.name == AppEvent.CTX_SELECTED:
                self.delayed_play(
                    self.window.core.access.helpers.get_selected_ctx(),  # with info about ctx
                    event,
                )
            elif event.name == AppEvent.MODE_SELECTED:
                self.delayed_play(
                    self.window.core.access.helpers.get_selected_mode(),  # with info about mode
                    event,
                )
            elif event.name == AppEvent.MODEL_SELECTED:
                self.delayed_play(
                    self.window.core.access.helpers.get_selected_model(),  # with info about model
                    event,
                )
            elif event.name == AppEvent.PRESET_SELECTED:
                self.delayed_play(
                    self.window.core.access.helpers.get_selected_preset(),  # with info about preset
                    event,
                )
            elif event.name == AppEvent.TAB_SELECTED:
                self.delayed_play(
                    self.window.core.access.helpers.get_selected_tab(),  # with info about tab
                    event,
                )
            elif event.name == AppEvent.CTX_END:
                if (not self.window.controller.plugins.is_type_enabled("audio.output") and
                        not self.window.controller.plugins.is_type_enabled("audio.control")):
                    self.delayed_play(
                        trans(trans_key),
                        event,
                    )  # only if audio output is disabled
            else:
                # handle rest of events
                self.delayed_play(
                    trans(trans_key),
                    event,
                )
            self.window.core.debug.info("AUDIO EVENT PLAY: " + event.name)

    def update(self):
        """Update voice control"""
        if self.window.core.config.get("access.voice_control"):
            self.window.ui.nodes['voice.control.btn'].setVisible(True)
        else:
            self.window.ui.nodes['voice.control.btn'].setVisible(False)

        self.window.controller.audio.update()  # update audio menu, etc.

    def enable_voice_control(self):
        """Enable voice control"""
        self.window.core.config.set("access.voice_control", True)
        self.window.core.config.save()
        self.update()

    def disable_voice_control(self):
        """Disable voice control"""
        self.window.core.config.set("access.voice_control", False)
        self.window.core.config.save()
        self.update()

    def toggle_voice_control(self):
        """Toggle voice control"""
        if self.is_voice_control_enabled():
            self.disable_voice_control()
        else:
            self.enable_voice_control()

    def is_voice_control_enabled(self) -> bool:
        """
        Check if voice control is enabled

        :return: True if enabled
        """
        return self.window.core.config.get("access.voice_control")

    def toggle_recording(self):
        """Toggle recording"""
        if self.is_recording:
            self.stop_recording()
        else:
            self.start_recording()

    def switch_btn_stop(self):
        """Switch button to stop"""
        self.window.ui.nodes['voice.control.btn'].btn_toggle.setText(trans('audio.speak.btn.stop'))
        self.window.ui.nodes['voice.control.btn'].btn_toggle.setToolTip(
            trans('audio.speak.btn.stop.tooltip'))

    def switch_btn_start(self):
        """Switch button to start"""
        self.window.ui.nodes['voice.control.btn'].btn_toggle.setText(trans('audio.control.btn'))
        self.window.ui.nodes['voice.control.btn'].btn_toggle.setToolTip(trans('audio.speak.btn.tooltip'))

    def stop_timeout(self):
        """Stop timeout"""
        self.stop_recording(timeout=True)

    def start_recording(self):
        """Start recording"""
        # display snap warning if not displayed yet
        if (not self.window.core.config.get("audio.input.snap", False)
                or not self.window.core.config.has("audio.input.snap")):
            if self.window.core.platforms.is_snap():
                self.window.ui.dialogs.open(
                    'snap_audio_input',
                    width=400,
                    height=200
                )
                self.window.core.config.set("audio.input.snap", True)
                self.window.core.config.save()
                return
        try:
            self.is_recording = True
            self.switch_btn_stop()

            # stop audio output if playing
            if self.window.controller.audio.is_playing():
                self.window.controller.audio.stop_output()

            # set audio volume bar
            self.window.core.audio.capture.set_bar(
                self.window.ui.nodes['voice.control.btn'].bar
            )

            # start timeout timer to prevent infinite recording
            timeout = int(self.window.core.config.get('audio.input.timeout', 120) or 0)  # get timeout
            if timeout > 0:
                if self.timer is None:
                    self.timer = QTimer()
                    self.timer.timeout.connect(self.stop_timeout)
                    self.timer.start(timeout * 1000)

            if not self.window.core.audio.capture.check_audio_input():
                raise Exception("Audio input not working.")
                # IMPORTANT!!!!
                # Stop here if audio input not working!
                # This prevents the app from freezing when audio input is not working!

            self.window.core.audio.capture.start()  # start recording if audio is OK
            self.window.update_status(trans('audio.speak.now'))
            self.window.dispatch(AppEvent(AppEvent.VOICE_CONTROL_STARTED))  # app event
        except Exception as e:
            self.is_recording = False
            self.window.core.debug.log(e)
            self.window.ui.dialogs.alert(e)
            if self.window.core.platforms.is_snap():
                self.window.ui.dialogs.open(
                    'snap_audio_input',
                    width=400,
                    height=200,
                )
            self.switch_btn_start()  # switch button to start

    def stop_recording(self, timeout: bool = False):
        """
        Stop recording

        :param timeout: True if stopped due to timeout
        """
        self.window.core.audio.capture.reset_audio_level()
        self.is_recording = False
        if self.timer:
            self.timer.stop()
            self.timer = None
        self.switch_btn_start()  # switch button to start
        path = os.path.join(self.window.core.config.path, self.input_file)
        self.window.core.audio.capture.set_path(path)

        if self.window.core.audio.capture.has_source():
            self.window.core.audio.capture.stop()  # stop recording
            # abort if timeout
            if timeout:
                self.window.dispatch(AppEvent(AppEvent.VOICE_CONTROL_STOPPED))  # app event
                self.window.update_status("Aborted.".format(timeout))
                return

            if self.window.core.audio.capture.has_frames():
                frames = self.window.core.audio.capture.get_frames()
                if len(frames) < self.MIN_FRAMES:
                    self.window.update_status(trans("status.audio.too_short"))
                    self.window.dispatch(AppEvent(AppEvent.VOICE_CONTROL_STOPPED))  # app event
                    return
                self.window.dispatch(AppEvent(AppEvent.VOICE_CONTROL_SENT))  # app event
                self.handle_thread(True)  # handle transcription in simple mode
        else:
            self.window.update_status("")
            self.window.dispatch(AppEvent(AppEvent.VOICE_CONTROL_STOPPED))  # app event

    def handle_thread(self, force: bool = False):
        """
        Handle listener thread

        :param force: force start
        """
        if self.thread_started and not force:
            return

        try:
            # worker
            worker = ControlWorker()
            worker.window = self.window
            worker.path = os.path.join(
                self.window.core.config.path,
                self.input_file,
            )
            # signals
            worker.signals.transcribed.connect(self.handle_transcribed)
            worker.signals.finished.connect(self.handle_input)
            worker.signals.destroyed.connect(self.handle_destroy)
            worker.signals.started.connect(self.handle_started)
            worker.signals.stopped.connect(self.handle_stop)
            worker.signals.status.connect(self.handle_status)
            worker.signals.error.connect(self.handle_error)

            # start
            self.window.threadpool.start(worker)

        except Exception as e:
            self.window.core.debug.log(e)

    @Slot(str, str)
    def handle_transcribed(self, path: str, text: str):
        """
        Handle transcribed text

        :param path: audio file path
        :param text: transcribed text
        """
        self.window.tools.get("transcriber").on_transcribe(path, text)

    @Slot(object, object)
    def handle_input(self, text: str, ctx: Optional[CtxItem] = None):
        """
        Insert text to input and send

        :param text: text
        :param ctx: CtxItem
        """
        if text is None or text.strip() == '':
            self.window.update_status("")
            return
        self.window.core.debug.info("VOICE CONTROL INPUT: " + text)
        commands = self.window.core.access.voice.recognize_commands(text)
        self.handle_commands(commands)

    def handle_commands(self, commands: List[Dict[str, Any]]):
        """
        Handle commands

        :param commands: commands list
        """
        unrecognized = False
        if len(commands) > 0:
            for command in commands:
                cmd = command["cmd"]
                params = command.get("params", "")
                self.window.update_status(trans("event.audio.cmd").format(cmd=cmd))
                event = ControlEvent(cmd)
                event.data = {
                    "params": params,
                }
                if event.name == "unrecognized":
                    unrecognized = True
                self.window.dispatch(event)
                self.window.core.debug.info("VOICE CONTROL COMMAND: " + cmd, params)
                trans_key = "event.control." + cmd

                if cmd in self.confirm_events:
                    self.window.update_status(trans("event.audio.confirm"))
                else:
                    event_name = trans(trans_key)
                    if event_name == trans_key:  # if no translation
                        event_name = cmd
                    self.window.update_status(trans("event.audio.cmd").format(cmd=event_name))
                    QApplication.processEvents()

            # play OK sound
            if not unrecognized:
                if self.window.core.config.get("access.audio.notify.execute"):
                    self.window.controller.audio.play_sound("ok.mp3")

    @Slot(object)
    def handle_error(self, data: str):
        """
        Handle thread error signal

        :param data: message
        """
        self.window.update_status(str(data))

    @Slot(object)
    def handle_status(self, data: str):
        """
        Handle thread status msg signal

        :param data: message
        """
        self.window.update_status(str(data))

    @Slot()
    def handle_destroy(self):
        """Handle listener destroyed"""
        self.thread_started = False

    @Slot()
    def handle_started(self):
        """Handle listening started"""
        self.thread_started = True

    @Slot()
    def handle_stop(self):
        """Handle stop listening"""
        self.thread_started = False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.16 01:00:00                  #
# ================================================== #

from llama_index.core.readers.base import BaseReader

from .hub.yt.base import YoutubeTranscriptReader
from .base import BaseLoader


class Loader(BaseLoader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.id = "youtube"
        self.name = "YouTube"
        self.type = ["web"]
        self.instructions = [
            {
                "youtube": {
                    "description": "read and index YouTube video URL",
                    "args": {
                        "url": {
                            "type": "str",
                            "label": "Youtube URL",
                            "description": "URL of the YouTube video, e.g. https://www.youtube.com/watch?v=CRRlbK5w8AE",
                        },
                    },
                }
            }
        ]
        self.init_args = {
            "languages": ["en"],
        }
        self.init_args_types = {
            "languages": "list",
        }
        self.init_args_labels = {
            "languages": "Languages",
        }
        self.init_args_desc = {
            "languages": "List of languages to extract from the video, separated by comma (,), e.g. 'en,de,fr'. Default is 'en'",
        }

    def get(self) -> BaseReader:
        """
        Get reader instance

        :return: Data reader instance
        """
        args = self.get_args()
        return YoutubeTranscriptReader(**args)

    def prepare_args(self, **kwargs) -> dict:
        """
        Prepare arguments for reader

        :param kwargs: keyword arguments
        :return: args to pass to reader
        """
        args = {}
        args["ytlinks"] = [kwargs.get("url")]  # list of links
        return args

    def is_supported_attachment(self, source: str) -> bool:
        """
        Check if attachment is supported by loader

        :param source: attachment source
        :return: True if supported
        """
        yt_prefix = [
            "https://youtube.com",
            "https://youtu.be",
            "https://www.youtube.com",
            "https://m.youtube.com",
        ]
        for prefix in yt_prefix:
            if source.startswith(prefix):
                return True
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 18:00:00                  #
# ================================================== #

import os

from .base import BaseProvider


class OpenAITextToSpeech(BaseProvider):
    def __init__(self, *args, **kwargs):
        """
        OpenAI Text to Speech provider

        :param args: args
        :param kwargs: kwargs
        """
        super(OpenAITextToSpeech, self).__init__(*args, **kwargs)
        self.plugin = kwargs.get("plugin")
        self.id = "openai_tts"
        self.name = "OpenAI TTS"
        self.allowed_models = [
            'tts-1',
            'tts-1-hd',
        ]

    def init_options(self):
        """Initialize options"""
        self.plugin.add_option(
            "openai_model",
            type="text",
            value="tts-1",
            label="Model",
            tab="openai_tts",
            description="Specify model, available models: "
                        "tts-1, tts-1-hd",
        )
        self.plugin.add_option(
            "openai_voice",
            type="combo",
            value="alloy",
            label="Voice",
            tab="openai_tts",
            use="audio_tts_whisper_voices",
            description="Specify voice, available voices: "
                        "alloy, echo, fable, onyx, nova, shimmer",
        )

    def speech(self, text: str) -> str:
        """
        Speech text to audio

        :param text: text to speech
        :return: path to generated audio file or None if audio playback is handled here
        """
        client = self.plugin.window.core.gpt.get_client()
        output_file = self.plugin.output_file
        voice = self.plugin.get_option_value('openai_voice')
        model = self.plugin.get_option_value('openai_model')
        allowed_voices = self.plugin.window.core.audio.whisper.get_voices()
        if model not in self.allowed_models:
            model = 'tts-1'
        if voice not in allowed_voices:
            voice = 'alloy'
        path = os.path.join(
            self.plugin.window.core.config.path,
            output_file,
        )
        response = client.audio.speech.create(
            model=model,
            voice=voice,
            input=text,
        )
        response.stream_to_file(path)
        return path

    def is_configured(self) -> bool:
        """
        Check if provider is configured

        :return: True if configured, False otherwise
        """
        return True

    def get_config_message(self) -> str:
        """
        Return message to display when provider is not configured

        :return: message
        """
        return ""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.18 19:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        """
        Set default options for plugin

        :param plugin: plugin instance
        """
        keys = {
            "enabled": "bool",
            "name": "text",
            "instruction": "textarea",
            "get_params": "textarea",
            "post_params": "textarea",
            "post_json": "textarea",
            "headers": "textarea",
            "type": {
                "type": "combo",
                "keys": [
                    {"GET": "GET"},
                    {"POST": "POST"},
                    {"POST_JSON": "POST_JSON"},
                ],
            },
            "endpoint": "textarea",
        }
        items = [
            {
                "enabled": True,
                "name": "search_wiki",
                "instruction": "send API call to Wikipedia to search pages by query",
                "get_params": "query, limit",
                "post_params": "",
                "post_json": "",
                "headers": "",
                "type": "GET",
                "endpoint": 'https://en.wikipedia.org/w/api.php?action=opensearch&limit={limit}&format=json&search={query}',
            },
        ]
        desc = "Add your custom API calls here"
        tooltip = "See the documentation for more details about examples, usage and list of predefined placeholders"
        plugin.add_option(
            "cmds",
            type="dict",
            value=items,
            label="Your custom API calls",
            description=desc,
            tooltip=tooltip,
            keys=keys,
        )
        plugin.add_option(
            "disable_ssl",
            type="bool",
            value=False,
            label="Disable SSL verify",
            description="Disables SSL verification when making calls to API",
            tooltip="Disable SSL verify",
        )
        plugin.add_option(
            "timeout",
            type="int",
            value=5,
            label="Timeout",
            description="Connection timeout (seconds)",
            tooltip="Connection timeout (seconds)",
        )
        plugin.add_option(
            "user_agent",
            type="text",
            value="Mozilla/5.0",
            label="User agent",
            description="User agent to use when making requests, default: Mozilla/5.0",
            tooltip="User agent to use when making requests",
        )